#include <Client/AssetService.hpp>
#include <thread>
#include <JGL/types/Texture.h>
#include <JJX/JSON.hpp>

using namespace std::chrono_literals;

static CaveGame::Client::AssetService* singleton;

CaveGame::Client::AssetService * CaveGame::Client::AssetService::Get() { return singleton; }

CaveGame::Client::AssetService::AssetService() {
    singleton = this;

    ParseManifest();
}



std::string CaveGame::Client::AssetService::FilenameFromPath(const std::filesystem::path &path) {
    return path.string().substr(path.string().find_last_of("/\\") + 1);
}

std::string CaveGame::Client::AssetService::FilenameFromPathWithoutExtension(const std::filesystem::path &path) {
    auto base = path.string().substr(path.string().find_last_of("/\\") + 1);

    std::string::size_type const p(base.find_last_of('.'));
    return base.substr(0, p);
}

void CaveGame::Client::AssetService::EnqueueTexture(const std::string& name, const std::filesystem::path &path) {
    queue.push({name, path, AssetType::TEXTURE});
    total_queued++;
}

void CaveGame::Client::AssetService::EnqueueFont(const std::string& name, const std::filesystem::path &path) {
    queue.push({name, path, AssetType::FONT});
    total_queued++;
}

void CaveGame::Client::AssetService::EnqueueSound(const std::string& name, const std::filesystem::path &path) {
    queue.push({name, path, AssetType::AUDIO});
    total_queued++;
}

bool CaveGame::Client::AssetService::LoadAsset(const CaveGame::Client::AssetRequest &request) {
    last_asset_processed = request.path.string();

    switch(request.type) {
        case AssetType::TEXTURE: {
            if (textures.contains(request.name)) // TODO: Note repeat request.
                return true;

            auto texture = std::make_shared<Texture>(request.path, FilteringMode::NEAREST);
            textures[request.name] = texture;
        }
        default: {
            // TODO: We don't support this asset type yet!!!
        }
    }
    last_asset_processed = request.path.string(); //FilenameFromPath(request.path);
    total_loaded++;
    return true;
}

bool CaveGame::Client::AssetService::LoadFromQueue(float maxTimeExpenditure) {

    using fsec = std::chrono::duration<float>;
    using clock = std::chrono::high_resolution_clock;

    if (queue.empty())
        return true;


    auto start = clock::now();

    //while (!queue.empty() && maxTimeExpenditure > 0.f)
    //{
        AssetRequest request = queue.front();
        queue.pop();

        if (!LoadAsset(request))
        {
            std::cerr << "bruh" << std::endl;
            // Failed to load!!!
        }
        auto cur_time = clock::now();
        fsec fs = cur_time - start;
        maxTimeExpenditure -= fs.count();
        //std::this_thread::sleep_for(250ms);
    //}
    return queue.empty();

}

void CaveGame::Client::AssetService::LoadAllFromQueue() {

    while (!queue.empty())
    {
        AssetRequest request = queue.front();
        queue.pop();
        if (!LoadAsset(request))
        {
            std::cerr << "bruh" << std::endl;
            // Failed to load!!!
        }
    }
}

void CaveGame::Client::AssetService::ParseManifest() {
    std::string contents = read_file("assets/data/manifest.json");

    using namespace JJX;

    auto [obj, errcode] = json::parse(contents);

    json::object catalog = obj.as_object();

    // TODO: in this case, calling obj.at("textures") results in a bad_alloc?

    if (catalog.contains("textures")) {
        json::array texlist = catalog.at("textures").as_array();
        for (auto& texture_entry : texlist) {
            std::string name = texture_entry[0].string.value();
            std::string path = texture_entry[1].string.value();
            EnqueueTexture(name, path);
        }
    }

    if (catalog.contains("fonts")) {
        for (auto &font_entry: catalog["fonts"].as_array()) {
            std::string name = font_entry[0].string.value();
            std::string path = font_entry[1].string.value();
            EnqueueFont(name, path);
        }
    }

    if (catalog.contains("sfx")) {
        for (auto &sfx_entry: catalog["sfx"].as_array()) {
            std::string name = sfx_entry[0].string.value();
            std::string path = sfx_entry[1].string.value();
            EnqueueSound(name, path);
        }
    }

    if (catalog.contains("music")) {
        for (auto &sfx_entry: catalog["music"].as_array()) {
            std::string name = sfx_entry[0].string.value();
            std::string path = sfx_entry[1].string.value();
            EnqueueSound(name, path);
        }
    }
}

void CaveGame::Client::AssetService::PreloadCertainAssets() {
    LoadAsset({"redacted", "assets/textures/redacted.png", AssetType::TEXTURE});
    LoadAsset({"title", "assets/textures/title_1.png", AssetType::TEXTURE});
}
