#include <format>
#include <Client/LocalWorld.hpp>
#include <JGL/JGL.h>
#include <Core/Data.hpp>
#include <fstream>
#include <Core/Loggers.hpp>
#include <JUI/Widgets/Scene.hpp>
#include <Core/TileRegistry.hpp>

namespace CaveGame::Client {
    using namespace CaveGame::Core;

    LocalWorld::LocalWorld(const std::string& world_name, int seed, bool overwrite)
        : World(world_name, seed, overwrite) {
        Logs::Info("Spinning up ChunkServerThread");
        chunk_thread = std::thread(&LocalWorld::ChunkServerThread, this);
        chunk_thread.detach();
        //hud = nullptr;
    }

    void LocalWorld::PostInit() {
        //font = JGL::Font("assets/fonts/Jupiteroid.ttf");
    }



    void LocalWorld::CheckCachedChunkSprites() {
        for (auto& [chunk_pos, chunk] : loaded_chunks) {
            if (!IsChunkCellWithinViewport(chunk_pos))
                continue;

            auto it = cached_chunk_sprites.find(chunk_pos);
            if (it != cached_chunk_sprites.end()) {
                if (!chunk->touched)
                    continue;

                chunk->touched = false;
                RenderChunkTexture(chunk_pos, it->second, chunk);
            }

            auto* chunk_sprite = new RenderTarget(Vector2i(Chunk::ChunkSize), Colors::Transparent,false,
                                                  SampleRate::NONE, FilteringMode::MIPMAP_NEAREST);
            RenderChunkTexture(chunk_pos, chunk_sprite, chunk);
            cached_chunk_sprites.insert({chunk_pos, chunk_sprite});
        }
    }

    Color4 LocalWorld::GetSkyColorInterpolatedForTimeOfDay(float time) {
        float rounded_low = Math::Floor(time);
        float rounded_high = Math::Ceil(time);

        float delta = (time - rounded_low);

        Color4 color_low = GetSkyColorBaseForTimeOfDay(rounded_low);
        Color4 color_high = GetSkyColorBaseForTimeOfDay(rounded_high);

        return Color4::Lerp(color_low, color_high, delta);
    }
    
    Color4 LocalWorld::GetSkyColorBaseForTimeOfDay(float time) {
        if (time >= 23*60)
            return Colors::Black;
        else if (time >= 22*60)
            return Colors::Black;
        else if (time >= 21*60)
            return Colors::White;
        else if (time >= 20*60)
            return Colors::White;
        else if (time >= 19*60)
            return Colors::White;
        else if (time >= 18*60)
            return Colors::White;
        else if (time >= 17*60)
            return Colors::White;
        else if (time >= 16*60)
            return Colors::White;
        else if (time >= 15*60)
            return Colors::White;

        return Colors::White;
    }

    void LocalWorld::DrawSky() {

        Vector2 viewport_topleft = camera.ScaledViewport().minPoint;
        Vector2 viewport_bottomright = camera.ScaledViewport().maxPoint;

        Color4 base_bg_color = Colors::Blues::CornflowerBlue;

        float space_starts_at = -2000.f;

        float depth = camera.Position().y;

        // Draw space bg.
        if (depth < -space_starts_at) {
            float t = Math::Clamp(  -((camera.Position().y+space_starts_at) / space_starts_at), 0.f, 1.f);
            base_bg_color = base_bg_color.Lerp(Colors::Blues::MidnightBlue, t);
        }

        float deep_starts_at = 2000.f;

        // Draw deep bg.
        if (depth > deep_starts_at) {
            float t = Math::Clamp((camera.Position().y-deep_starts_at) / deep_starts_at, 0.f, 1.f);
            base_bg_color = base_bg_color.Lerp(Colors::Black, t);
        }


        glClearColor(base_bg_color.RN(), base_bg_color.GN(), base_bg_color.BN(),1.f);

        //if ()

    }
    /*
    void LocalWorld::DrawBeforeProjection(){ }
    void LocalWorld::DrawDuringProjection() { }
    void LocalWorld::DrawAfterProjection() { }
     */

    void LocalWorld::Draw() {
        CheckCachedChunkSprites();

        DrawSky();

        J2D::Begin();

        // TODO: The following Translation, Rotation, Scale transformation code is duplicated between here and CaveGameWindow.cpp:Draw.
        // THIS BELONGS IN THE CAMERA ITSELF. You "Render" the camera first, And everything else after such that the camera can observe it. - Redacted.

        // Shift the origin to the center of the screen.
        glTranslatef(camera.HalfSizeOffset().x, camera.HalfSizeOffset().y, 0);

        //DrawSky();

        // Apply rotation, zoom, and translation.
        glRotatef(camera.Rotation(), 0, 0, 1);
        glScalef(camera.Zoom(), camera.Zoom(), 1);

        glTranslatef((int64_t) -camera.Position().x, (int64_t) -camera.Position().y, 0);

        // Draw the cached RenderTargets for our chunks.
        for (const auto& [chunk_pos, chunk] : loaded_chunks)
        {
            if (IsChunkCellWithinViewport(chunk_pos))
            {
                RenderChunk(chunk_pos);

                // Debug Grid
                if (chunk_debug.IsEnabled())
                {
                    chunk_debug.DrawStats(chunk, 0.99f / camera.Zoom());
                    chunk_debug.DrawCellCoords(chunk_pos, 0.99f / camera.Zoom());
                }

            }
        }

        // Debug Grid
        if (chunk_debug.IsEnabled())
            chunk_debug.DrawGrid(camera.ScaledViewport());

        for (auto* entity : entities) {
            entity->Draw();
        }

        for (auto& p : particles)
        {
            if (p.life > 0)
                p.Draw();
        }


        JGL::J2D::End();


        //hud->Draw();
    }

    unsigned int LocalWorld::GetRenderTargetCount() const {
        return cached_chunk_sprites.size();
    }

    bool LocalWorld::IsChunkCellWithinViewport(const Vector2i &coords, int extraChunkRadius) const {
        int extraChunkComputedPixelsRequired = extraChunkRadius*Core::Chunk::ChunkSize;
        AABB2D chunk_bounding_box = AABB2D(Vector2(coords)*Core::Chunk::ChunkSize, Vector2(coords)*Core::Chunk::ChunkSize+Vector2(Core::Chunk::ChunkSize));
        return Core::Solver::AABB2Dvs(
                camera.ScaledViewportOversized(extraChunkComputedPixelsRequired), chunk_bounding_box);
    }

    RNG rng;


    void LocalWorld::RenderTile(const TileID& t_id, int wx, int wy, int tx, int ty)
    {

        using namespace CaveGame::Core;

        static Tile stone = Tiles()["stone"];

        static TileID air = Tiles()["air"].numeric_id;
        static TileID void_tile = Tiles()["void"].numeric_id;
        static TileID cobblestone = Tiles()["cobblestone"].numeric_id;
        static TileID stone_id = stone.numeric_id;
        static TileID oak_plank = Tiles()["oak-plank"].numeric_id;
        static TileID stone_brick = Tiles()["stone-brick"].numeric_id;



        // TODO: Migrate custom tile render code to an override Draw method in tile classes?
        // TODO: See class Tile::Draw to see why this is halted.

        if (t_id == air || t_id == void_tile)
            return;


        Color4 t_color;


        const Core::Tile& t_data = Tiles()[t_id];

        if (t_id == cobblestone) {
            float val = generator.Perlin(wx, wy, 4, 4, 0.25f, 1.5f);


            unsigned int rand = generator.ColorMap(stone.pallet->size(), wx, wy);
            t_color = t_data.pallet->operator[](rand);

            if (val > 0.40f || val < -0.40f) {
                t_color.r += 32;
                t_color.g += 32;
                t_color.b += 32;
            } else if (val > 0.15f || val < -0.10f)
            { } else {
                t_color.r -= 64;
                t_color.g -= 64;
                t_color.b -= 64;
            }


            //Core::Tiles::Cobblestone.color_pallet
        } else if (t_id == oak_plank) {

            // TODO: Make each plank have a slightly different base color.

            int plank_height = 4;

            int plank_row = ((int) Math::Floor(wy / plank_height));

            int plank_row_idx = plank_row % 5;

            int plank_length_modulus = 12;

            uint8_t base_r = 212;// - shift;
            uint8_t base_g = 184;// - shift;
            uint8_t base_b = 125;// - shift;


            if (plank_row_idx == 0) { plank_length_modulus = 14; }
            if (plank_row_idx == 1) { plank_length_modulus = 19; }
            if (plank_row_idx == 2) { plank_length_modulus = 21; }
            if (plank_row_idx == 3) { plank_length_modulus = 17; }
            if (plank_row_idx == 4) { plank_length_modulus = 27; }

            int plank_col = (int) Math::Floor(wx / plank_length_modulus);

            uint8_t plank_base_rng = generator.ColorMap(20, plank_col, plank_row);
            uint8_t rng = generator.ColorMap(25, wx, wy);

            base_r -= plank_base_rng + rng;
            base_g -= plank_base_rng + rng;
            base_b -= plank_base_rng + rng;


            if (Math::Abs(wy % plank_height) == (plank_height-1) || Math::Abs((wx + (plank_row * plank_height)) % plank_length_modulus) == 1) {
                base_r += 30;
                base_g += 25;
                base_b += 20;
            }

            if (wy % plank_height == 0 || (wx + (plank_row * plank_height)) % plank_length_modulus == 0) {
                base_r -= 60;
                base_g -= 45;
                base_b -= 30;
            }

            t_color = {base_r, base_g, base_b};
        } else if (t_id == Tiles()["stone-brick"].numeric_id) {


            uint8_t shift = generator.ColorMap(30, wx, wy);
            uint8_t base_r = 130 - shift;
            uint8_t base_g = 125 - shift;
            uint8_t base_b = 120 - shift;

            int brick_height = 5;
            int brick_length = 9;
            int brick_row = ((int)Math::Floor(wy / brick_height));

            int brick_offset = 0;

            if (brick_row % 2 == 0)
                brick_offset = 3;


            if (Math::Abs(wy) % brick_height == (brick_height-1) || (Math::Abs(wx) + brick_offset) % brick_length == 1)
            {
                base_r += 35;
                base_g += 35;
                base_b += 30;
            }

            if (wy % brick_height == 0 || (wx + brick_offset) % brick_length == 0) {
                base_r -= 55;
                base_g -= 60;
                base_b -= 55;
            }
            t_color = {base_r, base_g, base_b};
        } else if (t_data.pallet.has_value()) {
            unsigned int rand = generator.ColorMap(t_data.pallet->size(), wx, wy);
            t_color = t_data.pallet.value()[rand];
        } else {
            t_color = t_data.color;
        }

        JGL::J2D::DrawPoint(t_color, tx, ty);
    }

    void LocalWorld::RenderChunkTexture(const Vector2i &coords, JGL::RenderTarget* destination, Core::Chunk* chunk)
    {

#define DEBUG_TILE_UPDATES

        using CaveGame::Core::TileID;

        TileID t_id;
        Core::Tile t_data;

        JGL::J2D::Begin(destination, nullptr, true);
        for (int x = 0; x < Core::Chunk::ChunkSize; x++)
        {
            for (int y = 0; y < Core::Chunk::ChunkSize; y++)
            {
                t_id = chunk->GetTile(x, y);

                Vector2 relative_tile_coords = Vector2(x, y);
                const Vector2& tile_coords = relative_tile_coords;
                int wx = (coords.x*Core::Chunk::ChunkSize) + x;
                int wy = (coords.y*Core::Chunk::ChunkSize) + y;

                //if (t_id == TileID::AIR || t_id == TileID::VOID) // Air
                    //continue;

                RenderTile(t_id, wx, wy, x, y);

                if (show_tile_activity)
                {
                    if (chunk->GetTileUpdateFlag(x, y))
                        JGL::J2D::DrawPoint(Color4(255, 0, 0, 64), tile_coords);

                    if (chunk->GetTileUpdateBufferFlag(x, y))
                        JGL::J2D::DrawPoint(Color4(255, 0, 0, 64), tile_coords);
                }


                /*t_data = Core::GetByNumeric(t_id);

                if (t_id == TileID::COBBLESTONE) {
                    float val = generator.Perlin(wx, wy, 12, 12, 12, 2);

                    if (val > 0.60f || val < -0.60f)
                        t_color = Core::Tiles::Cobblestone.color_pallet[1];
                    else if (val > 0.40f || val < -0.40f)
                        t_color = Core::Tiles::Cobblestone.color_pallet[0];
                    else if (val > 0.2f || val < -0.2f)
                        t_color = Core::Tiles::Cobblestone.color_pallet[2];
                    else
                        t_color = Core::Tiles::Cobblestone.color_pallet[3];
                    //Core::Tiles::Cobblestone.color_pallet
                } else if (t_id == TileID::OAK_PLANK) {

                    if (Math::Mod(y, 3) == 0 || Math::Mod(x, 9) == 0) {
                        t_color = Colors::Browns::Brown;
                    } else
                        t_color = Colors::Browns::BurlyWood;

                } else if (t_data->has_color_pallet) {
                    uint rand = generator.ColorMap(t_data->color_pallet.size(), x, y);
                    t_color = t_data->color_pallet[rand];
                } else {
                    t_color = t_data->base_color;
                }

                JGL::J2D::DrawPoint(t_color, tile_coords);*/

            }
        }
        JGL::J2D::End();
    }

    Vector2i GetChunkRealCoordinates(const Vector2i& cell) {
        return cell * Core::Chunk::ChunkSize;
    }

    void LocalWorld::RenderChunk(const Vector2i& coords) {
        auto it = cached_chunk_sprites.find(coords);
        if (it != cached_chunk_sprites.end())
            JGL::J2D::DrawRenderTarget(it->second, Vector2(GetChunkRealCoordinates(coords)));
    }

    void LocalWorld::LookForChunksNeedUnloading() {
        for (auto it = loaded_chunks.begin(); it != loaded_chunks.end();)
        {
            const auto coords = it->first;
            if (!IsChunkCellWithinViewport(coords, 2)) {
                // TODO: Move off main thread.
                SaveChunkToFile(coords, it->second);

                delete it->second;
                loaded_chunks.erase(it++);

                auto sprite_it = cached_chunk_sprites.find(coords);
                if (sprite_it != cached_chunk_sprites.end())
                    delete sprite_it->second, cached_chunk_sprites.erase(sprite_it);
            }
            else
                ++it;
        }
    }

    //void LocalWorld::LookForChunksNeedLoading()
    void LocalWorld::LookForChunksNeedLoading()
    {
        Vector2 viewport_topleft = camera.ScaledViewportOversized(2*Core::Chunk::ChunkSize).minPoint;
        Vector2 viewport_bottomright = camera.ScaledViewportOversized(2*Core::Chunk::ChunkSize).maxPoint;

        int lower_bound_h = Math::Floor(viewport_topleft.x / Core::Chunk::ChunkSize)/*-2*/;
        int upper_bound_h = Math::Floor(viewport_bottomright.x / Core::Chunk::ChunkSize)/*+2*/;

        int lower_bound_v = Math::Floor(viewport_topleft.y / Core::Chunk::ChunkSize)/*-2*/;
        int upper_bound_v = Math::Floor(viewport_bottomright.y / Core::Chunk::ChunkSize)/*+2*/;

        for (int x = lower_bound_h; x <= upper_bound_h; x++)
        {
            for (int y = lower_bound_v; y <= upper_bound_v; y++)
            {
                Vector2i cell(x, y);
                if (!HasChunkAtCell(cell) && !AwaitingChunkAtCell(cell))
                    RequestChunk(cell);
            }
        }
    }

    void LocalWorld::Update(float elapsed) {
        World::Update(elapsed);

        camera.Update(elapsed);


        for (auto& particle : particles)
        {
            if (particle.life > 0)
                particle.Update(elapsed);
        }

        std::erase_if(particles, [&](Particle& p) {return p.life <= 0; });

        /*std::vector<Particle>::iterator it;
        for (it = particles.begin(); it != particles.end();)
        {aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
            if (it->life <= 0)
                it = particles.erase(it);
            //else
                //it->Update(elapsed);
        }*/

        check_chunks_timer += elapsed;

        // TODO: Move this block to World class

        if (check_chunks_timer > 1/8.f) {
            // TODO: These procedures need to be performed asynchronously.
            check_chunks_timer = 0.f;
            LookForChunksNeedUnloading();
            LookForChunksNeedLoading();
            //auto a1 = std::async(&LocalWorld::LookForChunksNeedLoading, this);

            // TODO: Unload chunks once they're offscreen for a certain time threshold.


            // Check our generator queue for complete chunks, and pull them into our loaded chunks.
            while (!ServedChunks.empty())
            {
                Core::Chunk* c = nullptr;
                ServedChunks.front_pop(c);
                loaded_chunks.emplace(c->GetChunkCell(), c);

                if (std::find(chunks_in_waiting.begin(), chunks_in_waiting.end(), c->GetChunkCell()) != chunks_in_waiting.end())
                    chunks_in_waiting.erase(std::remove(chunks_in_waiting.begin(), chunks_in_waiting.end(), c->GetChunkCell()), chunks_in_waiting.end());

                //generator.SecondPass(this, c);
            }
        }
    }




    void LocalWorld::RefreshAll() {
        //World::RefreshAll();
        cached_chunk_sprites.clear();
        chunks_in_waiting.clear();
    }

    void LocalWorld::DebugChunks(bool enabled) {
        chunk_debug.Enable(enabled);
    }

    void LocalWorld::SaveAndExit() {
        World::SaveAndExit();

    }

    void LocalWorld::SetShowTileActivity(bool enabled) {
        show_tile_activity = enabled;
    }

    bool LocalWorld::IsShowTileActivityEnabled() const { return show_tile_activity;}

    void LocalWorld::Emit(const Particle &p) {
        particles.push_back(p);
    }


}

