#include <Client/CreditsWindow.hpp>
#include "ClientApp/CaveGameWindow.hpp"
#include <Client/SettingsMenu.hpp>

#include <Core/Explosion.hpp>
#include <Core/Loggers.hpp>
#include <Core/Player.hpp>
#include <ReWindow/InputService.h>
#include <Core/Macros.hpp>
#include <jstick.hpp>
#include <Core/ItemRegistry.hpp>
#include <Core/TileRegistry.hpp>
#include <JJX/JSON.hpp>
#include <JUI/Widgets/FpsGraph.hpp>

namespace CaveGame::ClientApp {

    using namespace CaveGame::Core;
    using namespace CaveGame::Client;

    void ReadRecipesAndRegister() {}

    CaveGameWindow::CaveGameWindow(const std::string& title, int width, int height): ReWindow::OpenGLWindow(title, width, height, 2, 1)
    {
        Logs::Info("Parsing Tile Data.");
        CaveGame::Core::LoadTileMetadata();
        Logs::Info("Parsing Item Data.");
        CaveGame::Core::LoadItemMetadata();
        Logs::Info("Creating game window.");

        CreateContexts();
        CreateMenuWindows();
    }

    CaveGameWindow::~CaveGameWindow() {

        Logs::Info("Removing game objects.");
        delete splash_ctx;
        delete menu_ctx;
        delete game_ctx;
        delete wm;

        Logs::Info("Closing game window.");
    }

    void CaveGameWindow::OpenWorld(Client::SingleplayerSessionInfo info)
    {
        Logs::Info(std::format("Game session requested.", info.NewWorld));
        game_ctx = new CaveGame::Client::GameSession(info.NewWorld);

        // TODO: Parse Info to construct gameworld files.
        ChangeScene(game_ctx);

        game_ctx->RequestToggleSettings += [&, this] {
            settings_window->Toggle();
        };

        game_ctx->OnSessionExit += [&, this] {
            ChangeScene(menu_ctx);
        };
    }

    void CaveGameWindow::CreateContexts()
    {
        Logs::Info("Building splash screen.");
        splash_ctx = new CaveGame::Client::Splash();
        Logs::Info("Building main menu.");
        menu_ctx = new CaveGame::Client::MainMenu();

        menu_ctx->RequestWorld += [this](Client::SingleplayerSessionInfo info) {
            OpenWorld(info);
        };
        menu_ctx->RequestQuit += [this] { Die(); };
        menu_ctx->RequestToggleCredits  += [this]{ credits_window->Toggle(); };
        menu_ctx->RequestToggleSettings += [this] { settings_window->Toggle(); };

        game_ctx = new CaveGame::Client::GameSession();
    }

    void CaveGameWindow::CreateMenuWindows()
    {
        wm = new JUI::Scene();
    }

    void CaveGameWindow::Step() {

        auto begin = GetTimestamp();
        Refresh();
        auto possible_end = GetTimestamp();

        float dt = ComputeElapsedFrameTimeSeconds(begin, possible_end);

        if (this->max_fps > 0) {
            float min_delta = 1.f / max_fps;
            float delta_diff = min_delta - dt;
            // TODO: only guaranteed to sleep *at least* as long as we request!
            // Fake bias to sleep for less time than we want, so it averages out?
            //std::this_thread::sleep_for(std::chrono::microseconds((long long)(delta_diff*1000*500)));

        }
        auto sure_end = GetTimestamp();
        dt = ComputeElapsedFrameTimeSeconds(begin, sure_end);
        UpdateFrameTiming(dt);
    }

    void CaveGameWindow::Gameloop()
    {
        while (!this->wanna_die) {
            Step();
        }
    }

    void CaveGameWindow::Run()
    {
        int js = 0;//jstick::Connect();

        //this->SetRenderer(RenderingAPI::OPENGL);
        bool success = this->Open();

        if (!success)
        {
            Logs::Info("Error initializing!!");
            return;
        }

        this->Init();
        this->SetResizable(true);
        this->SetVsyncEnabled(false);

        this->assets.PreloadCertainAssets();
        this->assets.ParseManifest();

        //for (int i = 0; i < 10; i++)
        //{
            //assets.EnqueueTexture("assets/textures/redacted.png");
            //assets.EnqueueTexture("assets/textures/bg.png");
            //assets.EnqueueTexture("assets/textures/player.png");
            //assets.EnqueueTexture("assets/textures/explosion.png");
            //assets.EnqueueTexture("assets/textures/title_1.png");
            //assets.EnqueueTexture("assets/textures/ash_wip_potions.png");
        //}


        ChangeScene(splash_ctx);
        //OpenWorld({.NewWorld = false});
        //this->ChangeScene(this->game_ctx);

        // TODO: Implement Resource/Asset manager rather than passing asset references around like this.
        this->menu_ctx->BuildWidgets();

        Gameloop();

        processOnClose();
        Cleanup();

        //DestroyWindow();
    }

    void CaveGameWindow::create_stats_window()
    {

    }

    void CaveGameWindow::create_settings_window()
    {
        settings_window = new Client::SettingsMenu(wm);
    }


    void CaveGameWindow::create_window_widgets()
    {
        create_console_window();
        credits_window = Client::CreateCreditsWindowWidget(wm);
        create_stats_window();
        create_settings_window();


        // TODO: Swap out with FpsGraphWindow on next JUI release.
        auto* graph_window = new JUI::Window(wm);
        graph_window->Name("FPS Graph");
        graph_window->Title("FPS Graph");
        graph_window->Size({500_px, 50_px});

        auto* graph = new JUI::FpsGraph(graph_window->Content());
        graph->Name("Graph");
        graph->Size({500_px, 50_px});
    }

    void CaveGameWindow::Init()
    {
        // Load OpenGL
        gladLoadGL();

        // Init Josh Graphics
        auto size = GetSize();
        Vector2i vsize = Vector2i(size.x, size.y);
        bool result = JGL::Init(vsize, 0.f, 0.f);
        //JGL::InitTextEngine();
        JGL::Update(vsize);

        // Fetch core assets
        // TODO: Asset manager? Need to share references and pointers of assets between many files..

        // Setup base OpenGL state
        glClearColor(0.f, 0.f, 0.f, 0.f);
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LESS);
        glDepthMask(GL_TRUE);

        // TODO: Delete when we update to the next release of JGL
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1); // NOTE: This MUST be called for text rendering to work properly!!!

        // Construct menu elements
        create_window_widgets();
    }

    void CaveGameWindow::draw_debug_info(std::vector<std::string> tokens)
    {
        int font_size = 14;
        Color4 bg_color = {0,0,0,192};
        Color4 text_color = Colors::Blues::SkyBlue;
        Vector2 pos = {2, 2};
        Vector2 text_size = JGL::Fonts::Jupiteroid.MeasureString(tokens[0], font_size);

        JGL::J2D::Begin();
        int index = 0;

        for (auto token : tokens)
        {
            text_size = JGL::Fonts::Jupiteroid.MeasureString(token, font_size);
            JGL::J2D::FillRect(bg_color, pos, text_size);
            JGL::J2D::DrawString(text_color, token, pos.x, pos.y, font_size);
            pos.y += text_size.y;
        }
        JGL::J2D::End();
    }

    void CaveGameWindow::InGameControls(float elapsed)
    {
        // Tile placer tool code

        // Temp hack until Input Focus Priority is implemented.
        for (auto window : wm->GetChildren())
            if (window->IsVisible() && window->IsMouseInside())
                return;

        // NOTE: Don't use else conditions for input code, you'll inadvertently add a bias for one key over the other.

        // TODO: Nuke this once JUI can consume input.
        if (InGame())
            GameSession()->WorldEditToolControlsUpdate(elapsed);
    }

    void CaveGameWindow::Update(float elapsed)
    {
        jstick::ReadEventLoop();

        SceneManager::UpdateSceneState(elapsed);

        // Update floating windows.
        wm->Update(elapsed);

        // TODO: We can't tell the game to change scenes from within a scene, currently.
        if (current_scene == splash_ctx && splash_ctx->SplashComplete())
            ChangeScene(menu_ctx);

        // Update Current Scene
        if (current_scene != nullptr)
        {
            auto isize = GetSize();
            Vector2 size = Vector2(isize.x, isize.y);
            current_scene->PassWindowSize(size);
            current_scene->Update(elapsed);
            // TODO: current_scene->PassInputState(...); ?
        }


        if (current_scene == GameSession())
        {
            InGameControls(elapsed);
        }
    }


    void CaveGameWindow::Draw()
    {
        auto isize = GetSize();
        Vector2i size = Vector2i(isize.x, isize.y);

        JGL::Update(size);

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        // Draw Current Scene
        if (current_scene != nullptr)
            current_scene->Draw();


        // Draw floating windows.
        wm->Draw();

        std::vector<std::string> debug_lines = {
            "Re-CaveGame - Redacted Software",
            std::format("frame: {}", refresh_count),
            std::format("delta: {}ms", J3ML::Math::Round(delta_time*1000)),
            std::format("fps: {}", J3ML::Math::Round(refresh_rate)),
            std::format("avg: {}", J3ML::Math::Round(avg_refresh_rate))
        };


        if (current_scene == GameSession()) {
            u16 id = game_ctx->GetTileIDUnderMouse();
            Core::Tile data = Tiles().Get(id);

            debug_lines.push_back(std::format("tile: {} id: {}", data.mnemonic_id, (unsigned int)data.numeric_id));
            debug_lines.push_back(std::format("entities: {}", game_ctx->World()->GetCurrentEntityCount()));
            debug_lines.push_back(std::format("tile_simulation: {}", game_ctx->World()->GetTileSimulationEnabled()));
            debug_lines.push_back(std::format("rendertargets: {}", game_ctx->World()->GetRenderTargetCount()));
        }

        draw_debug_info(debug_lines);
    }



    void CaveGameWindow::OnRefresh(float elapsed)
    {

        Update(elapsed);
        Draw();
        int glError = glGetError();
        if (glError != GL_NO_ERROR) {
            Logs::Error(std::format("OpenGL Error: {}", glError));
        }
        SwapBuffers();

    }

    void CaveGameWindow::OnKeyUp(const ReWindow::KeyUpEvent& ev) {
        if (current_scene != nullptr)
            current_scene->PassKeyInput(ev.key, false);

        if (ev.key == Keys::G)
            generate_grid = !generate_grid;

        if (ev.key == Keys::V)
            render_grid = !render_grid;

        if (ev.key == Keys::Grave) {

            // TODO: Key Input Hierarchy: Determine and handle priority of where certain keys go.
            console_window->SetOpen(!console_window->IsOpen());
            return;
        }

        //RWindow::OnKeyUp(ev);

        wm->ObserveKeyInput(ev.key, false);
    }

    void CaveGameWindow::OnKeyDown(const ReWindow::KeyDownEvent &ev) {

        if (current_scene == GameSession()) {

            if (ev.key == Keys::F5)
                GameSession()->World()->RefreshAll();

            if (ev.key == Keys::F7)
                GameSession()->World()->SetTileSimulationEnabled(!GameSession()->World()->GetTileSimulationEnabled());

            if (ev.key == Keys::F6) {
                std::srand(std::time(nullptr));
                GameSession()->World()->SetSeed(std::rand());
            }

            if (ev.key == Keys::Escape) {
                // TODO: "Pause menu"

                if (Console()->IsOpen())
                    Console()->SetOpen(false);


                //GameSession()->SaveAndExit();
                // TODO: GameContext needs mechanism to **inform** the higher-level object that we are done.
                //ChangeScene(menu_ctx);
            }
        }

        if (ev.key == Keys::Grave)
            return;

        if (current_scene != nullptr)
            current_scene->PassKeyInput(ev.key, true);

        wm->ObserveKeyInput(ev.key, true);

    }

    void CaveGameWindow::OnMouseButtonDown(const ReWindow::MouseButtonDownEvent& ev)
    {

        if (current_scene != nullptr)
            current_scene->PassMouseInput(ev.Button.ButtonIndex, true);

        if (ev.Button == MouseButtons::Left)
            wm->ObserveMouseInput(JUI::MouseButton::Left, true);
        if (ev.Button == MouseButtons::Right)
            wm->ObserveMouseInput(JUI::MouseButton::Right, true);
        if (ev.Button == MouseButtons::Middle)
            wm->ObserveMouseInput(JUI::MouseButton::Middle, true);


        mbd = true;


        RWindow::OnMouseButtonDown(ev);
    }


    void CaveGameWindow::OnMouseButtonUp(const ReWindow::MouseButtonUpEvent& ev)
    {
        if (current_scene != nullptr)
            current_scene->PassMouseInput(ev.Button.ButtonIndex, false);

        if (ev.Button == MouseButtons::Left)
            wm->ObserveMouseInput(JUI::MouseButton::Left, false);
        if (ev.Button == MouseButtons::Right)
            wm->ObserveMouseInput(JUI::MouseButton::Right, false);
        if (ev.Button == MouseButtons::Middle)
            wm->ObserveMouseInput(JUI::MouseButton::Middle, false);


        mbd = false;

        RWindow::OnMouseButtonUp(ev);
    }

    void CaveGameWindow::OnMouseMove(const ReWindow::MouseMoveEvent& ev) {

        Vector2 pos = Vector2(ev.Position.x, ev.Position.y);

        if (current_scene != nullptr)
            current_scene->PassMouseMovement(pos);

        wm->ObserveMouseMovement(pos);

        RWindow::OnMouseMove(ev);
    }

    bool CaveGameWindow::OnResizeRequest(const ReWindow::WindowResizeRequestEvent& ev) {

        return true;
    }

    void CaveGameWindow::OnMouseWheel(const ReWindow::MouseWheelEvent &ev) {
        if (current_scene)
            current_scene->PassMouseWheel(ev.WheelMovement);

        // wm->ObserveMouseWheel(ev.WheelMovement);
    }

    void CaveGameWindow::Die() { wanna_die = true; }

    bool CaveGameWindow::InMainMenu() const {
        return (current_scene == menu_ctx);
    }

    void CaveGameWindow::Cleanup() {

    }

    void CaveGameWindow::OnClosing() {
        if (current_scene == GameSession())
        {
            GameSession()->SaveAndExit();
        }

        wanna_die = true;
    }

    using namespace std::placeholders;

    std::string str_tolower(const std::string& s)
    {
        std::string copy = s;
        std::transform(copy.begin(), copy.end(), copy.begin(), ::tolower);
        return copy;
    }





    bool CaveGameWindow::HasCommand(const std::string& command)
    {
        for (const Command& c : commands) {
            if (c.name == command)
                return true;
            for(const std::string& alias : c.aliases)
                if (alias == command)
                    return true;
        }
        return false;
    }

    Command CaveGameWindow::GetCommand(const std::string& command) {
        for (Command c : commands) {
            if (c.name == command)
                return c;

            for (auto& alias : c.aliases)
            {
                if (command == alias)
                    return c;
            }
        }
        // TODO: sensible error.
        return InvalidCommand;
    }


    void CaveGameWindow::OnConsoleCommandInput(const std::string& command_sequence) {


        auto tokens = string_split(command_sequence, ' ');
        if (tokens.empty())
            return;

        std::string command_str = str_tolower(tokens[0]);

        if (!HasCommand(command_str)) {
            console_window->Log("Invalid command!");
            return;
        }

        Command cmd = GetCommand(command_str);


        cmd.callback(tokens);

    }

    void CaveGameWindow::create_console_window() {
        console_window = new JUI::CommandLine(this->wm);

        console_window->Visible(false);

        CaveGame::Logs::Info.OnLog += [this](std::string m, Color4 c) {Log(m, c); };
        CaveGame::Logs::Debug.OnLog += [this](std::string m, Color4 c) {Log(m, c); };
        CaveGame::Logs::Warning.OnLog += [this](std::string m, Color4 c) {Log(m, c); };
        CaveGame::Logs::Error.OnLog += [this](std::string m, Color4 c) {Log(m, c); };
        CaveGame::Logs::Fatal.OnLog += [this](std::string m, Color4 c) {Log(m, c); };

        CaveGame::Logs::Client.OnLog += [this](std::string m, Color4 c) {Log(m, c); };
        CaveGame::Logs::Server.OnLog += [this](std::string m, Color4 c) {Log(m, c); };
        CaveGame::Logs::Generator.OnLog += [this](std::string m, Color4 c) {Log(m, c); };
        CaveGame::Logs::Lighting.OnLog += [this](std::string m, Color4 c) {Log(m, c); };


        //console_window->
        console_window->OnInput += [this] (const std::string& msg)
        {
            OnConsoleCommandInput(msg);
        };
    }

    bool CaveGameWindow::InGame() const {
        return current_scene == game_ctx;
    }

    Vector2 CaveGameWindow::GetMouseV2() const {
        auto coords = GetMouseCoordinates();
        return Vector2(coords.x, coords.y);
    }

    Vector2 CaveGameWindow::GetSizeV2() const {
        auto isize = GetSize();
        return Vector2(isize.x, isize.y);
    }

    void CaveGameWindow::Log(const std::string &msg, const Color4 &color) {
        console_window->Log(msg, color);
    }

    bool CaveGameWindow::HelpCommand(const CommandArgs &args) {

        return true;
    }

    bool CaveGameWindow::ListCommand(const CommandArgs &args) {
        for (Command command: commands)
        {
            Log(std::format("{}", command.name));
        }
        return true;
    }

    Client::GameSession *CaveGameWindow::GameSession() { return game_ctx; }

    JUI::Scene *CaveGameWindow::WindowManager() { return wm;}

    JUI::CommandLine *CaveGameWindow::Console() { return console_window; }

    bool CaveGameWindow::ToggleTileSim(const CommandArgs &args) {
        if (InGame())
        {
            GameSession()->World()->SetTileSimulationEnabled(!GameSession()->World()->GetTileSimulationEnabled());
            return true;
        }

        Log("ERROR: Not in game context!", Colors::Red);
        return false;
    }

    void CaveGameWindow::MarkReadyToClose(bool close) { wanna_die = close;}

    bool CaveGameWindow::TileListCmd(const CommandArgs &args) {

        // TODO: Tile iterator that returns vector of **valid** tiles.
        const auto& tile_list = Core::Tiles().GetTileArray();

        for (const Core::Tile& tile : tile_list) {
            if (tile.numeric_id == 0 && tile.mnemonic_id != "air")
                continue; // Empty tile?
            Log(std::format("ID: {}, Mnemonic: {}, DisplayName: {}", tile.numeric_id, tile.mnemonic_id, tile.display_name));

        }

        return true;
    }

    bool CaveGameWindow::ItemListCmd(const CommandArgs &args) {
        auto item_list = Items().GetItemMap();

        for (const auto& [name, item] : item_list) {
            //if (item != nullptr)
                Log(std::format("Mnemonic: {}", name));
        }

        return true;
    }

    bool CaveGameWindow::NoclipCmd(const CommandArgs& args) {
        Log("Bet");
        if (InGame()) {
            auto plr = GameSession()->GetLocalPlayerEntity();
            if (plr != nullptr)
                plr->SetNoclip(!plr->IsNoclip());
        }
        return true;
    }

    bool CaveGameWindow::FpsLimitCmd(const CommandArgs& args) {
        if (args.size() < 2) {
            Log(std::format("Current FPS Limit: {}fps", max_fps));
            return true;
        }


        int val = stoi(args[1]);
        std::cout << val << std::endl;
        max_fps = val;

        return true;
    }

    bool CaveGameWindow::ReadyToClose() const { return wanna_die; }


}


