/// CaveGame - A procedural 2D platformer sandbox.
/// Created by Josh O'Leary @ Redacted Software, 2020-2024
/// Contact: josh@redacted.cc
/// Contributors: william@redacted.cc maxi@redacted.cc
/// This work is dedicated to the public domain.

/// @file Perlin.hpp
/// @desc A well-rounded Perlin noise class implementation.
/// @edit 10/1/2024
/// @auth Josh O'Leary

#pragma once

#include <cstdint>
#include <vector>

#include <J3ML/Algorithm/RNG.hpp>
#include <array>

namespace CaveGame::Core
{
    using J3ML::Algorithm::RNG;

    /// A class that implements and provides the Perlin Noise algorithm with acceptable performance characteristics.
    /// @see https://en.wikipedia.org/wiki/Perlin_noise
    class PerlinNoise {

    public:
        /// Default constructor that does not initialize any members.
        PerlinNoise() = default;

        /// This constructor returns a PerlinNoise object initialized with a given random number generator.
        /// @see class J3ML::Algorithm::RNG.
        explicit PerlinNoise(RNG rng);
        /// This constructor returns a PerlinNoise object initialized with a given random seed.
        /// This seed is passed to the RNG object directly.
        explicit PerlinNoise(int seed);
        /// Returns a sample of the noise gradient at a given 3D coordinate.
        double Noise(double x, double y, double z);

    protected:
    private:
        void InitGradients();
        int Index(int ix, int iy, int iz);
        double Lattice(int ix, int iy, int iz, double fx, double fy, double fz);
        static double Lerp(double t, double value0, double value1);
        int Permutate(int x);
        double Smooth(double x);

        constexpr static const double Offset = 0.5;
        static const int GradientSizeTable = 256;

        double gradients[GradientSizeTable*3];
        RNG random;

        /* Borrowed from Darwyn Peachey (see references above).
         The gradient table is indexed with an XYZ triplet, which is first turned
         into a single random index using a lookup in this table. The table simply
         contains all numbers in [0..255] in random order. */

        /* Changed to std::array because we know the size beforehand.
         * This also provides a small speedup because it's extremely
         * likely that when a given entry "g" is retrieved in Lattice,
         * g+1 and g+2 will already have been cached when fetching "g" */
        std::array<u8, 256> permutations = {
                225, 155, 210, 108, 175, 199, 221, 144, 203, 116, 70, 213, 69, 158, 33, 252, 5,
                82, 173, 133, 222, 139, 174, 27, 9, 71, 90, 246, 75, 130, 91, 191, 169, 138, 2, 151, 194, 235, 81, 7, 25,
                113, 228, 159, 205, 253, 134, 142, 248, 65, 224, 217, 22, 121, 229, 63, 89, 103, 96, 104, 156, 17, 201, 129,
                36, 8, 165, 110, 237, 117, 231, 56, 132, 211, 152, 20, 181, 111, 239, 218, 170, 163, 51, 172, 157, 47, 80,
                212, 176, 250, 87, 49, 99, 242, 136, 189, 162, 115, 44, 43, 124, 94, 150, 16, 141, 247, 32, 10, 198, 223,
                255, 72, 53, 131, 84, 57, 220, 197, 58, 50, 208, 11, 241, 28, 3, 192, 62, 202, 18, 215, 153, 24, 76, 41, 15,
                179, 39, 46, 55, 6, 128, 167, 23, 188, 106, 34, 187, 140, 164, 73, 112, 182, 244, 195, 227, 13, 35, 77, 196,
                185, 26, 200, 226, 119, 31, 123, 168, 125, 249, 68, 183, 230, 177, 135, 160, 180, 12, 1, 243, 148, 102, 166,
                38, 238, 251, 37, 240, 126, 64, 74, 161, 40, 184, 149, 171, 178, 101, 66, 29, 59, 146, 61, 254, 107, 42, 86,
                154, 4, 236, 232, 120, 21, 233, 209, 45, 98, 193, 114, 78, 19, 206, 14, 118, 127, 48, 79, 147, 85, 30, 207,
                219, 54, 88, 234, 190, 122, 95, 67, 143, 109, 137, 214, 145, 93, 92, 100, 245, 0, 216, 186, 60, 83, 105, 97,
                204, 52
        };
    };
}