#pragma once

#include <cstdint>
#include <cstddef>
#include <string>

namespace CaveGame::Core
{
    /// Primitive Serialization Functions.

    /// A data container structure which keeps track of the current position.

    struct Buffer {
        size_t size;
        size_t index;
        uint8_t* data;
    };



    /// Converts a 16-bit unsigned integer into a floating point number. @see float_to_u16.
    float u16_to_float(uint16_t value);
    /// Converts a floating point number to a 16-bit unsigned integer. This in effect, drops the accuracy of the float in exchange for compression.
    uint16_t float_to_u16(float);

    /// Writes a 1-byte unsigned int to the buffer, in network-byte-order, and advances the index by 1.
    void write_u8(Buffer& buffer, uint8_t value);
    /// Writes a 2-byte unsigned int to the buffer, in network-byte-order, and advances the index by 2.
    void write_u16(Buffer& buffer, uint16_t value);
    /// Writes a 4-byte unsigned int to the buffer, in network-byte-order, and advances the index by 4.
    void write_u32(Buffer& buffer, uint32_t value);
    /// Writes a 8-byte unsigned int to the buffer, in network-byte-order, and advances the index by 8.
    void write_u64(Buffer& buffer, uint64_t value);

    /// Reads a 1-byte unsigned int from the buffer, from network-byte-order, and advances the index by 1.
    uint8_t read_u8(Buffer& buffer);
    /// Reads a 2-byte unsigned int from the buffer, from network-byte-order, and advances the index by 2.
    uint16_t read_u16(Buffer& buffer);
    /// Reads a 4-byte unsigned int from the buffer, from network-byte-order, and advances the index by 4.
    uint32_t read_u32(Buffer& buffer);
    /// Reads a 8-byte unsigned int from the buffer, from network-byte-order, and advances the index by 8.
    uint64_t read_u64(Buffer& buffer);

    /// Writes a 1-byte signed int to the buffer, in network-byte-order, and advances the index by 1.
    void write_s8(Buffer& buffer, int8_t value);
    /// Writes a 2-byte signed int to the buffer, in network-byte-order, and advances the index by 2.
    void write_s16(Buffer& buffer, int16_t value);
    /// Writes a 4-byte signed int to the buffer, in network-byte-order, and advances the index by 4.
    void write_s32(Buffer& buffer, int32_t value);
    /// Writes a 8-byte signed int to the buffer, in network-byte-order, and advances the index by 8.
    void write_s64(Buffer& buffer, int64_t value);

    /// Reads a 1-byte signed int from the buffer, from network-byte-order, and advances the index by 1.
    int8_t read_s8(Buffer& buffer);
    /// Reads a 2-byte signed int from the buffer, from network-byte-order, and advances the index by 2.
    int16_t read_s16(Buffer& buffer);
    /// Reads a 4-byte signed int from the buffer, from network-byte-order, and advances the index by 4.
    int32_t read_s32(Buffer& buffer);
    /// Reads a 8-byte signed int from the buffer, from network-byte-order, and advances the index by 8.
    int64_t read_s64(Buffer& buffer);

    /// Writes a half-precision float to the buffer, via reinterpreting as uint16_t, and advances the index by 2.
    void write_f16(Buffer& buffer, float value);
    /// Writes a 4-byte float to the buffer, via reinterpreting as uint32_t, and advances the index by 4.
    void write_f32(Buffer& buffer, float value);
    /// Writes a 8-byte float to the buffer, via reinterpreting as uint64_t, and advances the index by 8.
    void write_f64(Buffer& buffer, double value);

    /// Reads a half-precision float from the buffer, via reinterpreting from uint16_t, and advances the index by 2.
    float read_f16(Buffer& buffer);
    /// Reads a 4-byte float from the buffer, via reinterpreting from uint32_t, and advances the index by 4.
    float read_f32(Buffer& buffer);
    /// Reads a 8-byte float from the buffer, via reinterpreting from uint64_t, and advances the index by 8.
    double read_f64(Buffer& buffer);

    void write_string(Buffer& buffer, const char* value, unsigned int length);

    void write_string(Buffer& buffer, std::string value);

    std::string read_string(Buffer& buffer);


    class ByteBuffer
    {
        Buffer buff;
        /// Write a uint8_t.
        ByteBuffer& operator<<(uint8_t& val);
        /// Write a uint16_t.
        ByteBuffer& operator<<(uint16_t& val);
        /// Write a uint32_t.
        ByteBuffer& operator<<(uint32_t& val);
    };

}