/// CaveGame - A procedural 2D platformer sandbox.
/// Created by Josh O'Leary @ Redacted Software, 2020-2024
/// Contact: josh@redacted.cc
/// Contributors: william@redacted.cc maxi@redacted.cc
/// This work is dedicated to the public domain.

/// @file Singleton.hpp
/// @desc An abstract class template that creates a Singleton.
/// @edit 3/20/2025
/// @auth Josh O'Leary

#pragma once

/// Template class for creating a Meyers' Singleton.
/// One and only one instance of this class can exist, and is alive throughout the lifetime of the program.
/// https://vlsiuniverse.blogspot.com/2016/04/meyers-singleton-pattern.html
/// The instance is lazy-initialized, meaning that:
///     if it is not a compile-time constant, and does not have a constructor,
///     it is only created upon the first call to Instance().
/// As per C++11, this is considered thread-safe out of the box.
template <class T>
class Singleton
{
public:
    static T& Instance()
    {
        // Since it's a static scoped variable, if the class has already been created, it won't be created again.
        // And it **is** thread safe (C++11 or later),
        static T instance;
        return instance;
    }
protected:
    Singleton() = default;
    virtual ~Singleton() = default;
    Singleton(const Singleton&) = delete;            // Copy constructor
    Singleton(Singleton&&) = delete;                 // Move constructor
    Singleton& operator=(const Singleton&) = delete; // Copy assign
    Singleton& operator=(Singleton&&) = delete;      // Move assign
};