#pragma once

#include <cstdint>
#include <Color4.hpp>
#include <Colors.hpp>
#include <Core/Data.hpp>
#include <map>
#include <functional>
#include <Core/Interfaces.hpp>
#include <J3ML/LinearAlgebra/Vector2i.hpp>
#include <J3ML/J3ML.hpp>


namespace CaveGame::Core {
    using J3ML::LinearAlgebra::Vector2i;
    using TileID = u16;

    struct Tile;

    using ColorPallet = std::vector<Color4>;
    using TileTiccFunc = std::function<void(const Tile& data, ITileMap* world, int x, int y)>;

    void SandGravTiccFunc(const Tile& data, ITileMap* world, int x, int y);
    void LiquidSettleTiccFunc(const Tile& data, ITileMap* world, int x, int y);
    void GrassRandomTiccFunc(const Tile& data, ITileMap* world, int x, int y);
    void GrassForcedTiccFunc(const Tile& data, ITileMap* world, int x, int y);
    void MossRandomTiccFunc(const Tile& data, ITileMap* world, int x, int y);
    void MossForcedTiccFunc(const Tile& data, ITileMap* world, int x, int y);
    void VineRandomTiccFunc(const Tile& data, ITileMap* world, int x, int y);
    void VineForcedTiccFunc(const Tile& data, ITileMap* world, int x, int y);


    static std::map<std::string, TileTiccFunc> ticc_funcs;

    // TODO: Tile light filling algorithm.
    // TODO: Ambient light from surface / certain biomes.
    // TODO: Directional lights.
    // TODO: Entity lights.
    // TODO: Tinting when interact with certain tiles.
    // TODO: Research recursive flood-fill algorithm.

    struct light {
        uint8_t r;
        uint8_t g;
        uint8_t b;
        uint8_t intensity;

    };

    // TODO: Implement params for tile color interaction
    // * Opacity
    // * Illumination
    // * What if we want to make tiles reactive, glow under UV, absorb light energy?

    struct Tile {
        std::string mnemonic_id;
        std::string display_name;
        std::optional<TileID> assigned_numeric_id;
        TileID numeric_id;
        std::optional<ColorPallet> pallet;
        bool solid = true;
        bool collides = false;
        bool does_random_ticc = false;
        bool does_forced_ticc = false;
        Color4 color;
        TileTiccFunc forced_ticc_func;
        TileTiccFunc random_ticc_func;
        std::vector<std::string> tags;
        bool gravity;
        bool opaque = true;
        std::optional<Color3> illumination;

    };
    /*
     *
     *  Tile and TileTrait example structure
     *  - Maxine
     *
     */

    /*
    class TileTrait
    {
    public:
        virtual std::string MnemonicID() = 0;
        virtual void Invoke() = 0;
    };

    class Tile
    {
    public:
        virtual std::string MnemonicID() = 0;
        virtual std::vector<TileTrait> Traits() = 0;
        virtual void AddTrait(TileTrait t) = 0;
        virtual void DelTrait(TileTrait t) = 0;
    };

    class Spreadable : public TileTrait {
    public:
        std::string MnemonicID() { return "tr_spreadable"; };
        void Invoke() {
            // Spread to adjacent tiles using calculation
            calc(world, owner);
        };

    public:
        using Calculator = std::function<void(ITileMap *w, Tile* o)>;
        Spreadable(ITileMap* w, Tile* o, Calculator c) : TileTrait() {
            world = w;
            owner = o;
            calc = c;
        };

    private:
        ITileMap* world;
        Tile* owner;
        Calculator calc;
    };

    class DecayTo : public TileTrait {
    public:
        std::string MnemonicID() { return "tr_decayto"; };
        void Invoke() {
            // Decay to tile on condition
            if (cond(owner))
                perf(world, owner);
        };

    public:
        using Condition = std::function<bool(Tile* o)>;
        using Perform = std::function<void(ITileMap* w, Tile* o)>;
        DecayTo(ITileMap* w, Tile* o, Condition c, Perform p) : TileTrait() {
            world = w;
            owner = o;
            cond = c;
            perf = p;
        }

    private:
        ITileMap* world;
        Tile* owner;
        Condition cond;
        Perform perf;
    };
     */

    // Use this for space-efficiency and convert to Color3 for doing color manipulations.
    // Struct allocation is cheaper.
    struct Light3 {
        uint8_t r;
        uint8_t b;
        uint8_t g;
    };
    /*std::unordered_map<std::string, TileTiccFunc> tile_functions {
        {"check-spread", [](const Tile& data, ITileMap* world, int x, int y){

        }},
        {"check-decay", [](const Tile& data, ITileMap* world, int x, int y){}},
        {"check-decay-spread", [](const Tile& data, ITileMap* world, int x, int y){}},
        {"sand-ticc", [](const Tile& data, ITileMap* world, int x, int y){}},
        {"water-ticc", [](const Tile& data, ITileMap* world, int x, int y){}},
        {"liquid-ticc", [](const Tile& data, ITileMap* world, int x, int y)
        {

        }}
    };*/



    class OnCondition
    {
    public:
        OnCondition();
        virtual bool Condition(ITileMap* world, int x, int y) { return true; }
        virtual void Perform(ITileMap* world, int x, int y) {}
        virtual void Invoke(ITileMap* world, int x, int y) {
            if (Condition(world, x, y))
                Perform(world, x, y);
        }
    };

    /// Tiles are instantiated as static members in the Tiles namespace.
    /// Chunks store tiles by their numeric ID, which can be used to retrieve the tile
    /*class Tile
    {
    protected:
        TileID numeric_id = TileID::VOID;
        std::string mnemonic_id;
        std::string display_name = "";
    public:
        Color4 base_color;
        std::vector<Color4> color_pallet;
        bool collides;
        bool does_random_ticc;
        bool has_color_pallet = false;
        bool solid = true;
    public:
        Tile();
        Tile(TileID id, const std::string& name, const Color4& color);
        Tile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet);
        virtual void Draw(int wx, int wy, int tx, int ty)
        {
            // TODO: Calling JGL requires being in the client code.
            // Maybe have Client/Tile.cpp implementation?
            //JGL::J2D::DrawPoint()
        }

        [[nodiscard]] TileID NumericID() const;
        [[nodiscard]] std::string MnemonicID() const;
        std::string Name() const;
        virtual bool DoesRandomTicc() const { return false; }
        virtual bool DoesForcedTicc() const { return false;}
        virtual bool Solid() const { return false;}
        virtual void ForcedTicc(ITileMap* world, TileState state, int x, int y) {}
        virtual void RandomTicc(ITileMap* world, TileState state, int x, int y) {}
        virtual void DecayTo(ITileMap *world, TileState state, int x, int y, TileID TDecaysTo);
        virtual bool ShouldSpread(ITileMap* world, int x, int y, TileID spreads_to) const;
        virtual bool ShouldSuffocate(ITileMap* world, int x, int y) const;
        virtual bool DecayCheck(ITileMap* world, TileState state, int x, int y, TileID decays_to);
        virtual bool SpreadCheck(ITileMap* world, TileState state, int x, int y, TileID spreads_to) { return false;}
    };*/

    /*class LiquidTile : public Tile
    {
    public:
        LiquidTile();
        LiquidTile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet);
        LiquidTile(TileID numeric, const std::string& name, Color4 color);
        bool Solid() const override { return true;}
        bool DoesForcedTicc() const override { return true;}
        void ForcedTicc(ITileMap *world, TileState state, int x, int y) override {

            // Supremely schizophrenic water solving which still doesn't 100% work.
            // Please fix this...


            TileID below = world->GetTile(x, y+1);
            TileID above = world->GetTile(x, y-1);
            TileID left = world->GetTile(x-1, y);
            TileID right = world->GetTile(x+1, y);
            TileID lookaheadLeft = world->GetTile(x-2, y);
            TileID lookaheadRight = world->GetTile(x+2, y);


            bool rng_roll = rand() % 2 == 0;

            if (below == TileID::AIR) {

                if (world->GetTile(x, y+2) == TileID::AIR) {
                    world->SetTile(x, y+2, numeric_id);
                    world->SetTile(x, y, TileID::AIR);
                    return;
                }

                world->SetTile(x, y+1, numeric_id);
                world->SetTile(x, y, TileID::AIR);
                return;
            }


            //if (below == TileID::WATER) {
            /*
                if (world->GetTile(rng_roll ? (x + 1) : (x - 1), y) == TileID::AIR)
                {
                    if (world->GetTile(rng_roll ? (x + 2) : (x - 2), y) == TileID::AIR)
                    {
                        if (world->GetTile(rng_roll ? (x + 3) : (x - 3), y) == TileID::AIR)
                        {
                            world->SwapTile(x, y, rng_roll ? (x + 3) : (x - 3), y);
                            return;
                        }
                        world->SwapTile(x, y, rng_roll ? (x + 2) : (x - 2), y);
                        return;
                    }

                //if (world->GetTile(rng_roll ? (x + 1) : (x - 1), y) == TileID::AIR) {
                TileID rollright = rng_roll ? right : left;
                TileID rollleft = !rng_roll ? right : left;
                if (rollright == TileID::AIR) {
                    world->SwapTile(x, y, rng_roll ? (x + 1) : (x - 1), y);
                    return;
                    //} else if (world->GetTile(!rng_roll ? (x + 1) : (x - 1), y) == TileID::AIR)
                } else if (rollleft == TileID::AIR)
                {
                    world->SwapTile(x, y, !rng_roll ? (x + 1) : (x - 1), y);
                    return;
                }

            if ((lookaheadRight == right) && (right == TileID::AIR)) {
                world->SetTile(x +1 , y, TileID::WATER);
                world->SetTile(x +2 , y, TileID::WATER);
                return;
            }

            if ((lookaheadLeft == left) && (left == TileID::AIR)){
                world->SetTile(x -1 , y, TileID::WATER);
                world->SetTile(x -2 , y, TileID::WATER);
                return;
            }

            if (below == TileID::WATER && above == TileID::AIR) {
                //if (world->GetTile(x + 1, y) == TileID::AIR && world->GetTile(x + 2, y) == TileID::WATER) {
                if (right == TileID::AIR && lookaheadRight == TileID::WATER) {
                    world->SetTile(x + 1, y, rng_roll ? TileID::WATER : TileID::AIR);
                    return;
                }

                //if (world->GetTile(x - 1, y) == TileID::AIR && world->GetTile(x - 2, y) == TileID::WATER) {
                if (left == TileID::AIR && lookaheadLeft == TileID::WATER) {
                    world->SetTile(x - 1, y, rng_roll ? TileID::WATER : TileID::AIR);
                    return;
                }
            }




           // }

            /*if (world->GetTile(x-1, y) == TileID::AIR) {
                if (world->GetTile(x-2, y) == TileID::AIR) {
                    world->SetTile(x-2, y, numeric_id);
                    world->SetTile(x, y, TileID::AIR);
                    return;
                }
                world->SetTile(x-1, y, numeric_id);
                world->SetTile(x, y, TileID::AIR);
                return;
            }

            if (world->GetTile(x+1, y) == TileID::AIR) {

                if (world->GetTile(x+2, y) == TileID::AIR) {
                    world->SetTile(x+2, y, numeric_id);
                    world->SetTile(x, y, TileID::AIR);
                    return;
                }
                world->SetTile(x+1, y, numeric_id);
                world->SetTile(x, y, TileID::AIR);
                return;
            }

            if (world->GetTile(x+1, y+1) == TileID::AIR) {
                world->SetTile(x+1, y+1, numeric_id);
                world->SetTile(x, y, TileID::AIR);
                return;
            }

            if (world->GetTile(x-1, y+1) == TileID::AIR) {
                world->SetTile(x-1, y+1, numeric_id);
                world->SetTile(x, y, TileID::AIR);
                return;
            }

            //if (world->GetTile(x-1, y) == TileID::AIR && world->GetTile(x+1, y) == TileID::AIR) {
                //world->SetTile(x-1, y, numeric_id);
                //world->SetTile(x, y, TileID::AIR);
                //return;
           //}

            if (world->GetTile(x-1, y) == TileID::AIR) {
                world->SetTile(x-1, y, numeric_id);
                world->SetTile(x, y, TileID::AIR);
                return;
            }

            if (world->GetTile(x+1, y) == TileID::AIR) {
                world->SetTile(x+1, y, numeric_id);
                world->SetTile(x, y, TileID::AIR);
                return;
            }
        }

    };

    class WaterTile : public LiquidTile
    {
    public:
        WaterTile();
        WaterTile(TileID id, const std::string& name, const Color4& color);
        void ForcedTicc(ITileMap* world, TileState state, int x, int y) override {
            if (world->GetTile(x, y-1) == TileID::LAVA) {
                world->SetTile(x, y-1, TileID::OBSIDIAN);
                world->SetTile(x, y, TileID::AIR);
                return;
            }

            if (world->GetTile(x, y+1) == TileID::LAVA) {
                world->SetTile(x, y+1, TileID::OBSIDIAN);
                world->SetTile(x, y, TileID::AIR);
                return;
            }

            if (world->GetTile(x-1, y) == TileID::LAVA) {
                world->SetTile(x-1, y, TileID::OBSIDIAN);
                world->SetTile(x, y, TileID::AIR);
                return;
            }

            if (world->GetTile(x+1, y) == TileID::LAVA) {
                world->SetTile(x+1, y, TileID::OBSIDIAN);
                world->SetTile(x, y, TileID::AIR);
                return;
            }
            LiquidTile::ForcedTicc(world, state, x, y);
        };
        bool DoesForcedTicc() const override { return true;}
    };

    class SoilTile : public Tile
    {
    public:
        TileID decays_to;

        SoilTile();
        SoilTile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet, TileID decays_target);
        SoilTile(TileID numeric, const std::string& name, Color4 color, TileID decays_target);
        bool DoesRandomTicc() const override { return true; }
        void RandomTicc(ITileMap *world, TileState state, int x, int y) override;
        bool SpreadCheck(ITileMap* world, TileState state, int x, int y, TileID spreads_to) override;
    };


    class GrassyTile : public SoilTile {
    public:
        GrassyTile();
        GrassyTile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet, TileID decays_target);
        GrassyTile(TileID numeric, const std::string& name, Color4 color, TileID decays_target);
        GrassyTile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet);
        GrassyTile(TileID numeric, const std::string& name, Color4 color);
        void RandomTicc(ITileMap *world, TileState state, int x, int y) override;
    };

    class MossyTile : public SoilTile {
    public:
        MossyTile();
        MossyTile(TileID id, const std::string& name, const Color4& color, TileID decays_target);
        MossyTile(TileID id, const std::string &name, const Color4 &color, const std::vector<Color4> &pallet);
        MossyTile(TileID numeric, const std::string &name, Color4 color);
        void RandomTicc(ITileMap *world, TileState state, int x, int y) override;
    };

    class VineTile : public Tile {
    public:
        VineTile();
        VineTile(TileID id, const std::string& name, const Color4& color);
        VineTile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet);

        bool Solid() const override { return true;}
        bool DoesRandomTicc() const override { return true; }
        void ForcedTicc(ITileMap* world, TileState state, int x, int y) override;
        void RandomTicc(ITileMap* world, TileState state, int x, int y) override;
        bool DecayCheck(ITileMap* world, TileState state, int x, int y, TileID decays_to) override;
        bool ShouldSpread(ITileMap* world, int x, int y, TileID spreads_to) const override;
        bool SpreadCheck(ITileMap* world, TileState state, int x, int y, TileID spreads_to) override;
    };

    class GravityTile : public Tile {
    public:
        GravityTile();
        GravityTile(TileID id, const std::string& name, const Color4& color);
        GravityTile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet);

        inline bool DoesForcedTicc() const override { return true;}
        inline bool DoesRandomTicc() const override { return true;}

        inline void ForcedTicc(ITileMap* world, TileState state, int x, int y) override;
    };

    class SaplingTile : public Tile {
    public:

    };

    class GasTile : public Tile
    {

    };

    void DoGrassRandomTicc(ITileMap *world, TileState state, int x, int y);


#define TILE static const Tile


    /*namespace Tiles {
        using ID = Core::TileID;
        static const Tile Void       {ID::VOID, "Void", Colors::Transparent};

        static const Tile Air            {ID::AIR,       "Air",       Colors::Transparent};
        static const Tile Stone          {ID::STONE,     "Stone",     Colors::Grays::SlateGray, {Colors::Grays::SlateGray, Colors::Grays::LightSlateGray}};
        static const Tile Dirt           {ID::DIRT,      "Dirt",      Colors::Browns::Chocolate, {Colors::Browns::Chocolate, {210, 125, 30},{195, 105, 40}}};
        static const Tile Mud            {ID::MUD,       "Mud",       Colors::Browns::SaddleBrown};
        static const Tile Limestone      {ID::LIMESTONE,  "Limestone", Colors::Yellows::PaleGoldenrod};
        static const Tile Basalt         {ID::BASALT, "Basalt", Colors::Grays::DimGray};
        static const Tile Cobblestone    {ID::COBBLESTONE, "Cobblestone", Colors::Grays::SlateGray, {Colors::Grays::SlateGray, Colors::Grays::LightSlateGray, {64, 64, 64}, {92, 92, 92}}};
        static const MossyTile RedMoss        {ID::RED_MOSS, "Red Moss ", Colors::Reds::Crimson, ID::STONE};
        static const MossyTile BrownMoss      {ID::BROWN_MOSS, "Brown Moss", Colors::Browns::Brown, ID::STONE};
        static const MossyTile GreenMoss      {ID::GREEN_MOSS, "Green Moss", Colors::Greens::LimeGreen, ID::STONE};
        static const MossyTile LavaMoss       {ID::LAVA_MOSS, "Lava Moss", Colors::Reds::LightCoral, ID::STONE};
        static const Tile Granite        {ID::GRANITE, "Granite", Colors::Whites::AntiqueWhite};
        static const Tile Marble         {ID::MARBLE, "Marble", Colors::Whites::GhostWhite};
        static const GrassyTile Grass    {ID::GRASS, "Grass", Colors::Greens::LawnGreen,
            {{126, 252, 5}, {122, 238, 0}, {124, 248, 12}}};
        static const GrassyTile GlowyGrass     {ID::GLOWY_GRASS, "Glowy Grass", Colors::Blues::PowderBlue};

        static const VineTile Vine {ID::VINE, "Vine", Colors::Greens::ForestGreen};

        static const GravityTile Sand           {ID::SAND, "Sand", Colors::Yellows::PaleGoldenrod, {{238, 232, 170}, {232, 238, 160}, {218, 212, 175}}};

        static const Tile WhiteSand      {ID::WHITE_SAND, "White Sand", Colors::Whites::SeaShell};
        static const Tile RedSand        {ID::RED_SAND, "Red Sand", Colors::Reds::Firebrick};
        static const Tile BlackSand      {ID::BLACK_SAND, "Black Sand", Colors::Black};

        static const Tile Sandstone {ID::SANDSTONE, "Sandstone", Colors::Yellows::PaleGoldenrod};
        static const Tile WhiteSandstone {ID::WHITE_SANDSTONE, "White Sandstone", Colors::Whites::SeaShell};
        static const Tile RedSandstone{ID::RED_SANDSTONE, "Red Sandstone", Colors::Reds::Firebrick};
        static const Tile BlackSandstone {ID::BLACK_SANDSTONE, "Black Sandstone", Colors::Black};

        static const Tile GrayBrick {ID::STONE_BRICK, "Gray Brick", Colors::Grays::DimGray};

        static const Tile Ash  {ID::ASH, "Ash", Colors::Grays::DarkSlateGray};
        static const Tile Clay {ID::CLAY, "Clay", Colors::Browns::Brown, {{164, 42, 42}, {164, 38, 42}, {172, 42, 52}, {164, 58, 62}}};
        static const Tile Silt {ID::SILT, "Silt", Colors::Browns::SaddleBrown};
        static const Tile Snow {ID::SNOW, "Snow", Colors::Whites::MintCream};
        static const Tile Ice {ID::ICE, "Ice", Colors::Blues::LightSteelBlue};
        static const Tile Slush {ID::SLUSH, "Slush", Colors::Blues::LightBlue};


        static const Tile Cactus {ID::CACTUS, "Cactus", Colors::Greens::Olive};
        static const Tile CactusFlower {ID::BLOOMING_CACTUS, "Cactus Flower", Colors::Reds::Salmon};

        static const Tile Bamboo;
        static const Tile Sugarcane;
        static const Tile VineFlower;
        static const Tile FlowerStem;
        static const Tile RedFlowerPetal;
        static const Tile BlueFlowerPetal;
        static const Tile YellowFlowerPetal;
        static const Tile WhiteFlowerPetal;

        static const Tile CopperWire {ID::COPPER_WIRE, "Copper Wire", Colors::Oranges::DarkOrange};
        static const Tile GoldWire {ID::GOLD_WIRE, "Gold Wire", Colors::Oranges::Gold};
        static const Tile RedWire {ID::RED_WIRE, "Red Wire", Colors::Red};
        static const Tile GreenWire {ID::GREEN_WIRE, "Green Wire", Colors::Green};
        static const Tile YellowWire {ID::YELLOW_WIRE, "Yellow Wire", Colors::Yellow};
        static const Tile BlueWire {ID::BLUE_WIRE, "Blue Wire", Colors::Blue};

        static const Tile LED;
        static const Tile TNT;

        static const Tile Cloud {ID::CLOUD, "Cloud", Colors::Whites::Azure};
        static const Tile ThinIce;
        static const Tile RainCloud;
        static const Tile Cobweb;

        static const Tile Coal {ID::COAL_ORE, "Coal Ore", Colors::Black};


        // TODO: Add AluminiumOre
        // TODO: Add ChromiumOre
        // TODO: Add NickelOre
        // TODO: Add ArsenicOre
        // TODO: Add GalliumOre

        static const Tile CopperOre {ID::COPPER_ORE, "Copper Ore", Colors::Oranges::DarkOrange, {{255, 140, 0}, {234, 150, 3}, {241, 138, 5}}};
        static const Tile TinOre {ID::TIN_ORE, "Tin Ore", Colors::Grays::Gainsboro};
        static const Tile IronOre {ID::IRON_ORE, "Iron Ore", Colors::Pinks::LightPink};
        static const Tile LeadOre {ID::LEAD_ORE, "Lead Ore", Colors::Grays::DarkSlateGray};
        static const Tile SilverOre {ID::SILVER_ORE, "Silver Ore", Colors::Grays::Silver};
        static const Tile TungstenOre {ID::TUNGSTEN_ORE, "Tungsten Ore", Colors::Grays::DimGray};
        static const Tile GoldOre {ID::GOLD_ORE, "Gold Ore", Colors::Oranges::Gold};
        static const Tile PlatinumOre {ID::PLATINUM_ORE, "Platinum Ore", Colors::Grays::Silver};
        static const Tile CobaltOre {ID::COBALT_ORE, "Cobalt Ore", Colors::Blues::MediumSlateBlue};
        static const Tile TitaniumOre {ID::TITANIUM_ORE, "Titanium Ore", Colors::Grays::DimGray};
        static const Tile UraniumOre {ID::URANIUM_ORE, "Uranium Ore", Colors::Greens::GreenYellow};

        static const Tile Diamond;
        static const Tile Emerald;
        static const Tile Amethyst;
        static const Tile Topaz;
        static const Tile Sapphire;
        static const Tile Ruby;
        static const Tile Amber;

        static const Tile Obsidian {ID::OBSIDIAN, "Obsidian", Colors::Purples::DarkMagenta};


        static const SaplingTile OakSapling;
        static const Tile OakLog {ID::OAK_LOG, "Oak Log", Colors::Browns::BurlyWood};
        static const Tile OakPlank {ID::OAK_PLANK, "Oak Plank", Colors::Browns::BurlyWood};
        static const Tile OakPlatform;
        static const Tile OakLeaf;

        static const Tile AshLog;
        static const Tile AshLeaf;
        static const Tile AshPlank;

        static const Tile PalmLog;
        static const Tile PalmLeaf;
        static const Tile PalmPlank;

        static const Tile PineLog;
        static const Tile PinePlank;
        static const Tile PinePlatform;
        static const Tile PineLeaf;

        static const Tile RedwoodLog;
        static const Tile RedwoodPlank;
        static const Tile RedwoodPlatform;
        static const Tile RedwoodLeaf;

        static const Tile EbonyLog;
        static const Tile EbonyPlank;
        static const Tile EbonyPlatform;
        static const Tile EbonyLeaf;

        static const Tile MapleLog;
        static const Tile MapleLeaf;
        static const Tile MaplePlank;


        static const Tile BirchLog;
        static const Tile BirchLeaf;
        static const Tile BirchPlank;


        static const WaterTile Water {ID::WATER, "Water", Colors::Blues::DodgerBlue};
        static const LiquidTile Blood;
        static const LiquidTile Sludge;
        static const LiquidTile Lava {ID::LAVA, "Lava", Colors::Red};
        static const LiquidTile MuddyWater;
        static const LiquidTile Ectoplasm;
        static const LiquidTile Oil;
        static const LiquidTile Honey;
        static const LiquidTile Milk;

        static const Tile Rope;

    }*/

}
