#include <map>
#include "Chunk.hpp"
#include "Generator.hpp"
#include "J3ML/LinearAlgebra/Vector2i.hpp"
#include "Interfaces.hpp"
#include <unordered_map>
#include <filesystem>
#include <future>
#include "ConcurrentQueue.hpp"
#include "Entity.hpp"
#include "Core/v2i_hash.hpp"

#pragma once



namespace CaveGame::Core {
    class WorldAsyncIOThreadManager
    {

    };

    /// The world class manages and coordinates terrain generation, chunk file IO, thread synchronization, and entity logic.
    class World : public ITileMap {
    public:
        Event<> OnNextDay;
        Event<> OnAutosave;

        constexpr static unsigned int RandomTileTickCoefficient = 128;
        constexpr static float PeriodicAutosaveIntervalSeconds = 30.f;
        constexpr static float PhysicsTickrate = 100.f; // Desired physics Ticks per second.
        World() = default;
        ~World() = default;

        explicit World(const std::string &worldName, int seed = 0, bool overwrite = false);


        Vector2i GetChunkCoordinatesAtCell(int x, int y) const {
            float chunkX = Math::Floor((float)x / Chunk::ChunkSize);
            float chunkY = Math::Floor((float)y / Chunk::ChunkSize);

            return Vector2i(chunkX, chunkY);
        }

        Vector2i GetTileCoordinatesAtCell(int x, int y) const {
            float tileX = Math::Mod(x, Chunk::ChunkSize);
            float tileY = Math::Mod(y, Chunk::ChunkSize);

            if (tileX < 0)
                tileX = Chunk::ChunkSize + tileX;
            if (tileY < 0)
                tileY = Chunk::ChunkSize + tileY;

            return Vector2i(tileX, tileY);
        }

        /// Returns the chunk-coordinate for the given set of tile coordinates.
        /// @param x world-space tile row.
        /// @param y world-space tile column.
        //Vector2i GetChunkCoordinatesAtCell(int x, int y) const;

        /// Returns the local-space tile coordinates [0,32] from the given set of world-space tile coordinates.
        /// @param x world-space tile row.
        /// @param y world-space tile column.
        //Vector2i GetTileCoordinatesAtCell(int x, int y) const;

        /// Returns the numeric ID of the tile a the given world-space coordinates.
        TileID GetTile(int x, int y) const override;

        /// Sets the tile stored at the given location.
        void SetTile(int x, int y, TileID t, bool flag_update = true) override;

        /// Swaps two tiles at given locations in-place.
        void SwapTile(int source_x, int source_y, int dest_x, int dest_y, bool flag_update = true) override;

        /// Returns whether tile simulation logic is enabled.
        bool GetTileSimulationEnabled() const;

        /// Set whether tile simulation logic is enabled.
        void SetTileSimulationEnabled(bool enabled);

        /// Returns the World Seed.
        int GetSeed() const;

        /// Sets the World Seed.
        void SetSeed(int seed);


        TileState GetTileState(int x, int y) const override { return 0; /* TODO: Implement tile state field */ }

        void SetTileState(int x, int y, TileState state) override { /* TODO: Implement tile state field */ }

        void SetTileUpdateFlag(int x, int y, bool flag) override;

        bool GetTileUpdateFlag(int x, int y) const override;

        // TODO: Doesn't really belong here.
        Vector2 ToUnitDirection(float rotation);

        Chunk* GetChunkAtCell(const Vector2i &cell);

        std::unordered_map<Vector2i, Chunk*> GetChunkList();


        void DoRandomTileTick(const Vector2i& coords, Chunk* chunk);

        void DoForcedTileTick(const Vector2i& coords, Chunk* chunk);

        void DoTileTiccs(float elapsed);

        virtual void Update(float elapsed);

        virtual void RefreshAll();


        void Save();
        void SaveAsync()
        {

        }

        /// Gracefully saves all game state and closes the play session.
        virtual void SaveAndExit();

        /// Returns whether or not the world currently has a chunk loaded at the specified chunk coordinates.
        /// @note Valid Chunk coordinates are integers only.
        bool HasChunkAtCell(const Vector2i &cell) const;

        bool HasChunkOnFile(const Vector2i &cell) const;

        bool AwaitingChunkAtCell(const Vector2i &cell) const;

        void ChunkServerThread();

        std::filesystem::path GetWorldRootPath() const;

        std::filesystem::path GetChunkFullPath(const Vector2i &cell) const;


        /// Adds the entity pointer to the world, via std::move, this may leave the original entity pointer empty.
        void AddEntity(Entity* e);

        std::vector<Entity*> GetEntities() { return entities;}


        /// Returns the number of extant entities in this world right now.
        unsigned int GetCurrentEntityCount() const;

        /// Marks that the chunk at the given cell coordinates is needed.
        /// This will be picked up by the generator thread, and either generated, or loaded from disk.
        void RequestChunk(const Vector2i& cell);
        void DropChunk(const Vector2i& cell);

        void SetTileTiccRate(float ticcrate);
        float GetTileTiccRate() const { return tile_ticc_frequency; }

    protected:
        void SaveChunkToFile(const Vector2i &cell, Chunk* chunk);
        virtual void AutoSaveAsync() {
            OnAutosave.Invoke();
            Logs::Info("Autosaving...");

            Save();
            Logs::Info("Autosave successful!");
        }
        virtual void AutoSave();

    protected:
        std::vector<Entity*> entities;
        const std::filesystem::path worlds {"worlds"};
        std::unordered_map<Vector2i, Chunk*> loaded_chunks;
        Generator generator;
        float time_of_day;
        std::string world_name;
        RNG rng;
        float autosave_timer = 0.f;
        bool simulate_tiles = true;
        std::thread chunk_thread;
        std::atomic<bool> run_chunk_thread = true;
        ConcurrentQueue<Vector2i> RequestedChunks;
        std::vector<Vector2i> chunks_in_waiting;
        ConcurrentQueue<Core::Chunk*> ServedChunks;

        float tile_ticc_counter = 0;
        float physics_ticc = 0;

        // Ticcs per second.
        float tile_ticc_frequency = 24.f;


        bool save_in_progress = false;
    private:


        void SimulateTiles(float elapsed);
    };
}