#include <Core/Generator.hpp>
#include <Core/Data.hpp>
#include <Core/TileRegistry.hpp>

namespace CaveGame::Core
{
    Generator::Generator(int seed) : perlin(seed), rng(seed)
    {
        this->seed = seed;
        simplex = SimplexNoise();

        for (int x = 0; x < PrecomputedWhiteNoiseResolution; x++)
        {
            precomputed_white_noise_1D[x] = rng.Float01Incl();

            for (int y = 0; y < PrecomputedWhiteNoiseResolution; y++)
            {
                precomputed_white_noise_2D[x][y] = rng.Float01Incl();
            }
        }
    }

    float Generator::ComputeDepth(int wx, int wy)
    {

        // this looks almost the same as the octave noise but with significantly less expense?
        //float height_map_low_pass = perlin.Noise(wx / HeightMapHighPassScale, 0.f, 0.25f) * HeightMapHighPassAmplitude;

        // TODO Use threading to run the high pass and low pass at the same time because it's expensive.
        float height_map_low_pass = perlin.Noise(3, wx / HeightMapHighPassScale, 0.f)* HeightMapHighPassAmplitude;
        float height_map_high_pass = perlin.Noise(wx / HeightMapLowPassScale, 0.f, 0.5f)* HeightMapLowPassAmplitude;


        float height_map = height_map_low_pass + height_map_high_pass;
        return wy - (BaseSurfaceLvl+height_map);
    }

    TileID Generator::HeightMap(float depth, int wx, int wy) {

        float rand = GetPrecomputedWhiteNoise2D(wx, wy)*(TopSoilDepth/2.f);

        if (wy > 1000)
            return Tiles()["stone"].numeric_id;

        if (depth < 0)  { return Tiles().MnemonicToNumeric("air"); }
        else if (depth < 2.5f) { return Tiles().MnemonicToNumeric("grass");; }
        else if (depth < TopSoilDepth-rand) { return Tiles().MnemonicToNumeric("dirt"); }
        else { return Tiles().MnemonicToNumeric("stone");; }
    }

    float Generator::Perlin(int wx, int wy, float hScale, float vScale, float offset, float outputScale) {
        return perlin.Noise(wx / hScale, wy / vScale, offset) * outputScale;
    }

    float Generator::Octave(int wx, int wy, float hScale, float vScale, float offset, float outputScale, int octaves) {

        float noise = 0;
        float frequency = 1;
        float factor = 1;
        float roughness = 3.f;
        float persistance = 0.4f;

        for (int i = 0; i < octaves; i++) {
            noise = noise + perlin.Noise(wx / hScale * (frequency + i) * 0.72354, wy / vScale * (frequency + i) * 0.72354, offset * (frequency + i) * 0.72354) * factor;
            factor *= persistance;
            frequency *= roughness;
        }
        return noise;
    }



    float Generator::AddSumAndPartialProduct(float hi, float lo, float div) {
        return hi + lo + ((hi * lo) / div);
    }

    TileID Generator::HiLoSelect(float pass, float upperBound, float lowerBound, TileID hiSelect, TileID loSelect, TileID fallback) {
        if (pass > upperBound)
            return hiSelect;

        if (pass < lowerBound)
            return loSelect;

        return fallback;
    }

    //bool Generator::CaveTest() { }


    float Generator::ComputeOre(int wx, int wy, const Vector2& hiPass, const Vector2& loPass, float hiOffset, float loOffset, float hiScale, float loScale, float noise, float ramp)
    {
        float pass_1 = Perlin(wx, wy, hiPass.x, hiPass.y, hiOffset, hiScale);
        float pass_2 = Perlin(wx, wy, loPass.x, loPass.y, loOffset, loScale);
        float rng = GetPrecomputedWhiteNoise2D(wx, wy)*noise;
        return AddSumAndPartialProduct(pass_1, pass_2, ramp) - rng;
    }

    //float Generator::ComputeLiquidPool() {}


    /// Calculates the tile to be placed at a given x,y coordinates during the first terrain iteration.
    TileID Generator::ComputeTile(int wx, int wy) {

        // TODO: Pre-generate height maps instead of computing perlin noise samples at runtime.

        float depth = ComputeDepth(wx, wy);
        TileID base = HeightMap(depth, wx, wy);

        float cave_erosion = Perlin(wx, wy, CaveInputScale, CaveInputScale, 0.f, CaveOutputScale);

        float rng = GetPrecomputedWhiteNoise2D(wx, wy)*CaveWhiteNoise;

        //float cave_addition = perlin.Noise(wx / CaveAdditiveInputScale, wy / CaveAdditiveInputScale, 0.f)*CaveAdditiveOutputScale;
        float cave_addition = Perlin(wx, wy, CaveAdditiveInputScale, CaveAdditiveInputScale, 0.25f, CaveAdditiveOutputScale);
        float cave_addition_2 = Perlin(wx, wy, 25, 25, 0.625f, 0.125f);
        cave_erosion += cave_addition;
        cave_erosion += cave_addition_2;
        cave_erosion -= rng;

        // Make cave openings more narrow
        if (depth < SurfaceCaveShrinkDepth) {
            cave_erosion *= SurfaceCaveShrinkFactor;
        }

        if (cave_erosion > -CaveErosionRange && CaveErosionRange > cave_erosion) {
            return Tiles()["air"].numeric_id;
        }


        if (depth > 1) {
            // Ore computation

            // Chunks of clay and dirt.
            float clay_pass = ComputeOre(wx, wy, ClayVeinHiPassScale, ClayVeinLoPassScale, ClayVeinHiPassOffset, ClayVeinLoPassOffset,
                ClayVeinHiPassOutputScale, ClayVeinLoPassOutputScale, ClayVeinNoise, ClayVeinRampFactor);

            // Yes, sometimes we re-use a pass to generate another tile entirely. Why?
            // We don't need the full parameter space for clay generation, as it's rarer, and it gives
            // more variety to dirt blob generation, for example.
            if (clay_pass > 0.85f) { return Tiles()["clay"].numeric_id; }
            if (clay_pass < -0.75f) { return Tiles()["dirt"].numeric_id; }

            // Chunks of silt and dirt.
            float silt_pass = ComputeOre(wx, wy, SiltVeinHiPassScale, SiltVeinLoPassScale, SiltVeinHiPassOffset, SiltVeinLoPassOffset,
                SiltVeinHiPassOutputScale, SiltVeinLoPassOutputScale, SiltVeinNoise, SiltVeinRampFactor);

            if (silt_pass > 0.85f) { return Tiles()["gravel"].numeric_id; }
            if (silt_pass < -0.85f) { return Tiles()["gravel"].numeric_id; }

            // Chunks of stone.
            float stone_pass = ComputeOre(wx, wy, StoneVeinHiPassScale, StoneVeinLoPassScale, StoneVeinHiPassOffset, StoneVeinLoPassOffset,
                StoneVeinHiPassOutputScale, StoneVeinLoPassOutputScale, StoneVeinNoise, StoneVeinRampFactor);

            // Chunks of stone
            if (stone_pass > 0.7f) { return Tiles()["stone"].numeric_id; }
            if (stone_pass < -0.7f) { return Tiles()["stone"].numeric_id; }

            // Chunks of dirt
            float dirt_pass = ComputeOre(wx, wy, DirtVeinHiPassScale, DirtVeinLoPassScale, DirtVeinHiPassOffset, DirtVeinLoPassOffset,
                DirtVeinHiPassOutputScale, DirtVeinLoPassOutputScale, DirtVeinNoise, DirtVeinRampFactor);

            if (-0.65 > dirt_pass && dirt_pass < 0.65f) { return Tiles()["dirt"].numeric_id; }

            // Coal veins
            float coal_pass = ComputeOre(wx, wy, CoalVeinHiPassScale, CoalVeinLoPassScale, CoalVeinHiPassOffset, CoalVeinLoPassOffset,
                CoalVeinHiPassOutputScale, CoalVeinLoPassOutputScale, CoalVeinNoise, CoalVeinRampFactor);

            if (coal_pass > 0.75f) { return Tiles()["coal-ore"].numeric_id; }
            if (coal_pass < -0.75f) { return Tiles()["coal-ore"].numeric_id;}

            // Copper ore veins
            float copper_pass = ComputeOre(wx, wy, CopperVeinHiPassScale, CopperVeinLoPassScale, CopperVeinHiPassOffset, CopperVeinLoPassOffset,
                CopperVeinHiPassOutputScale, CopperVeinLoPassOutputScale, CopperVeinNoise, CopperVeinRampFactor);

            if (copper_pass > 0.8f) { return Tiles()["copper-ore"].numeric_id; }

            // Tin ore veins
            float tin_pass = ComputeOre(wx, wy, TinVeinHiPassScale, TinVeinLoPassScale, TinVeinHiPassOffset, TinVeinLoPassOffset,
                TinVeinHiPassOutputScale, TinVeinLoPassOutputScale, TinVeinNoise, TinVeinRampFactor);

            if (tin_pass > 0.85f) { return Tiles()["tin-ore"].numeric_id; }

            // Iron ore veins
            float iron_pass = ComputeOre(wx, wy, IronVeinHiPassScale, IronVeinLoPassScale, IronVeinHiPassOffset, IronVeinLoPassOffset,
                IronVeinHiPassOutputScale, IronVeinLoPassOutputScale, IronVeinNoise, IronVeinRampFactor);

            if (iron_pass > 0.90f) { return Tiles()["iron-ore"].numeric_id;}

            if (depth > 1000) {
                // Lead ore veins

                // Silver ore veins

                // Tungsten ore veins

                // Gold ore veins
            }

            if (depth > 10000) {
                // Platinum ore veins

                // Cobalt ore veins

                // Titanium ore veins

                // Uranium ore veins
            }






             // This is actually faster??? - maxine
             // The compiler must be doing some optimization for all the function calls
             // It feels faster, but maybe I'm mistaken
             // I have also discovered that vines are major source of slowness possibly
           /*return HiLoSelect(
                       ComputeHiLoPass(
                       ComputePass(wx, wy, ClayVeinHiPassHScale, ClayVeinHiPassVScale, ClayVeinHiPassOffset, ClayVeinHiPassOutputScale),
                       ComputePass(wx, wy, ClayVeinLoPassHScale, ClayVeinLoPassVScale, ClayVeinLoPassOffset, ClayVeinLoPassOutputScale),
                       2.f) - rng, 0.85f, -0.75f, TileID::CLAY, TileID::DIRT,
               HiLoSelect(
                       ComputeHiLoPass(
                       ComputePass(wx, wy, 30.f, 30.f, 666.f, 2.4f),
                       ComputePass(wx, wy, 220.f, 220.f, 0.5f, 1.75f),
                       2.f) - rng, 0.85f, -0.75f, TileID::SILT, TileID::DIRT,
               HiLoSelect(
                       ComputeHiLoPass(
                       ComputePass(wx, wy, 30.f, 30.f, 666.f, 2.4f),
                       ComputePass(wx, wy, 220.f, 220.f, 0.5f, 1.75f),
                       1.75) + (rng*1.5f) , 0.7f, -0.7f, TileID::STONE, TileID::STONE, base
               )));*/
        }

        return base;
    }

    void Generator::FirstPass(CaveGame::Core::Chunk* chunk) {
        Vector2i real_coords = chunk->GetChunkRealCoordinates();

        for (int x = 0; x < Chunk::ChunkSize; x++) {
            for (int y = 0; y < Chunk::ChunkSize; y++) {

                int wx = real_coords.x + x;
                int wy = real_coords.y + y;

                chunk->SetTile(x, y, ComputeTile(wx, wy));
            }
        }

        chunk->SetFirstPassCompleted(true);
    }

    void Generator::SecondPass(ITileMap *world, Chunk *chunk) {
        Vector2i real_coords = chunk->GetChunkRealCoordinates();

        for (int x = 0; x < Chunk::ChunkSize; x++) {
            for (int y = 0; y < Chunk::ChunkSize; y++) {

                int wx = real_coords.x + x;
                int wy = real_coords.y + y;

                float tree_grove_noise = perlin.Noise(wx / 94.f, 0.f, 0.f);

                if (tree_grove_noise < 0.75f)
                    continue;


                // Check for tree placement.
                if (GetPrecomputedWhiteNoise1D(wx) > 0.5f)
                    continue;


                if (chunk->GetTile(x, y) != Tiles()["grass"].numeric_id) {}
                    continue;

                if (world->GetTile(wx, wy-1) != Tiles()["air"].numeric_id || world->GetTile(wx, wy-2) != Tiles()["air"].numeric_id   || world->GetTile(wx, wy-3) != Tiles()["air"].numeric_id )
                    continue;

                int treeHeight = 5 + (GetPrecomputedWhiteNoise1D(wx) * 12);

                for (int tree_y = 0; tree_y < treeHeight; tree_y++) {
                    world->SetTile(wx, wy-tree_y, Tiles()["oak-log"].numeric_id);
                }
            }
        }

        chunk->SetSecondPassCompleted(true);
    }

    unsigned int Generator::ColorMap(int range, int wx, int wy) {
        wx = ModInt(wx, PrecomputedWhiteNoiseResolution);
        wy = ModInt(wy, PrecomputedWhiteNoiseResolution);
        float norm = precomputed_white_noise_2D[wx][wy];
        return norm*range;
    }

    float Generator::GetPrecomputedWhiteNoise2D(int x, int y) const {
        x = ModInt(x, PrecomputedWhiteNoiseResolution);
        y = ModInt(y, PrecomputedWhiteNoiseResolution);
        return precomputed_white_noise_2D[x][y];
    }

    float Generator::GetPrecomputedWhiteNoise1D(int x) const {
        x = ModInt(x, PrecomputedWhiteNoiseResolution);
        return precomputed_white_noise_1D[x];
    }

}
