#include <Core/TileRegistry.hpp>
#include <Core/Loggers.hpp>
#include <format>
#include <Core/Item.hpp>
#include <Core/ItemRegistry.hpp>
#include <Core/Macros.hpp>
#include <Core/JsonConversions.hpp>
#include <JJX/JSON.hpp>

namespace CaveGame::Core
{
    void TileRegistry::Register(Tile data) {
        // If an ID is hardcoded to this tile, try to assign it.

        if (data.assigned_numeric_id.has_value()) {
            // Jump our current_index to this point
            current_index = data.assigned_numeric_id.value();

            // If this ID is consumed, scream an error.
            if (consumed_ids.at(current_index))
                throw std::runtime_error("ID is in use!!");

        }

        // If no ID is hardcoded, assign one from an increasing index.
        data.numeric_id = current_index;

        Register(current_index, data);

        current_index++;

    }

    void TileRegistry::Register(TileID ID, const Tile& data)
    {
        Logs::Info(std::format("\tRegister Tile:: id:{} mnemonic:{} display-name:{}", data.numeric_id, data.mnemonic_id, data.display_name));

        registered_tiles[ID] = data;

        // We also cache LUTs for fast access of name and integer ID from each other
        id_to_name_mapping[ID] = data.mnemonic_id;
        name_to_id_mapping[data.mnemonic_id] = ID;
        consumed_ids[ID] = true;
    }

    uint16_t TileRegistry::MnemonicToNumeric(const std::string &mnemonic) {
        return name_to_id_mapping[mnemonic];
    }

    const Tile &TileRegistry::Get(const std::string &mnemonic) {
        if (!name_to_id_mapping.contains(mnemonic))
            throw std::runtime_error(std::format("No tile found named {}", mnemonic));
        u16 id = name_to_id_mapping[mnemonic];

        return registered_tiles[id];
    }

    const Tile &TileRegistry::Get(u16 id) {
        return registered_tiles[id];
    }

    const Tile &TileRegistry::operator[](u16 ID) {
        return GetByNumericID(ID);
    }

    const Tile &TileRegistry::operator[](const std::string &mnemonic) {
        return GetByMnemonicID(mnemonic);
    }

    std::string TileRegistry::NumericToMnemonic(uint16_t ID) {
        return id_to_name_mapping[ID];
    }

    const Tile &TileRegistry::GetByNumericID(uint16_t ID) {
        return registered_tiles[ID];
    }

    const Tile &TileRegistry::GetByMnemonicID(const std::string &mnemonic) {
        return registered_tiles[MnemonicToNumeric(mnemonic)];
    }

    std::vector<Tile> TileRegistry::GetTileList() {
        return {std::begin(registered_tiles), std::end(registered_tiles)};
    }

    const std::array<Tile, TileRegistry::MaxTiles> &TileRegistry::GetTileArray() { return registered_tiles; }

    bool TileRegistry::Exists(const std::string &mnemonic) { return name_to_id_mapping.contains(mnemonic); }

    bool TileRegistry::Exists(u16 id) { return id_to_name_mapping.contains(id); }

    bool LoadTileMetadata(const std::filesystem::path &path) {

        // TODO: Move to it's own thing.
        ticc_funcs.insert({"sand-grav", SandGravTiccFunc});
        ticc_funcs.insert({"liquid-settle", LiquidSettleTiccFunc});
        ticc_funcs.insert({"grass-random", GrassRandomTiccFunc});
        ticc_funcs.insert({"grass-forced", GrassForcedTiccFunc});
        ticc_funcs.insert({"moss-random", MossRandomTiccFunc});
        ticc_funcs.insert({"moss-forced", MossForcedTiccFunc});
        ticc_funcs.insert({"vine-random", VineRandomTiccFunc});
        ticc_funcs.insert({"vine-forced", VineForcedTiccFunc});
        //ticc_funcs.insert({"water-solver", })

        // Subgroups for what collides with what. ~Will
        // Like collision mask? Or tile-reactions (i.e. obsidian and water)

        // Solid means the tile is noncompressible. It is a solid box.
        // Collides means whether any collision test is even performed. Will be used for cosmetic tiles. Terrain decoration.
        // Otherwise, i'm thinking you could use an idiom like
        // Tile collided;
        // if (collided.tagged("soil")) { ... }


        using namespace JJX;
        std::string content = read_file(path);

        auto [data, parse_error] = json::parse(content);

        if (data.type == json::value_type::array) {
            for (auto& te : data.as_array()) {
                Core::Tile tile;
                Core::Item tile_item;

                json::object tile_entry = te.as_object();

                if (tile_entry.contains("hardcoded-id"))
                    tile.assigned_numeric_id = tile_entry["hardcoded-id"].number;

                tile.mnemonic_id = tile_entry.at("mnemonic-id").string.value();
                tile.display_name = tile_entry.at("display-name").string.value();

                tile.solid = true;
                if (tile_entry.contains("solid"))
                    tile.solid = tile_entry.at("solid").boolean.value();

                tile_item.mnemonic = std::format("{}-tile", tile.mnemonic_id);
                tile_item.display_name = tile.display_name;

                tile.color = JsonConversions::parse_color(tile_entry["color"]);


                std::string ftfkey = "forced-ticc-func";


                // Retrieve and evaluate the force ticc function ID. Assign to tile if valid.
                tile.does_forced_ticc = false;
                if (tile_entry.contains(ftfkey)
                    && tile_entry[ftfkey].type == json::value_type::string
                    && !tile_entry[ftfkey].string.value().empty())
                {
                    std::string ticc_func_name = tile_entry[ftfkey].string.value();
                    if (!ticc_funcs.contains(ticc_func_name)) {
                        std::cerr << "No Tile Ticc Function with ID " << ticc_func_name << " found!" << std::endl;
                    } else {
                        tile.does_forced_ticc = true;
                        tile.forced_ticc_func = ticc_funcs.at(ticc_func_name);
                    }
                }

                // Retrieve and evaluate the random ticc function ID. Assign to tile if valid.
                std::string rtfkey = "random-ticc-func";
                tile.does_random_ticc = false;

                if (tile_entry.contains(rtfkey)
                    && tile_entry[rtfkey].type == json::value_type::string
                    && !tile_entry[rtfkey].string.value().empty())
                {
                    std::string ticc_func_name = tile_entry[rtfkey].string.value();
                    if (!ticc_funcs.contains(ticc_func_name)) {
                        std::cerr << "No Tile Ticc Function with ID " << ticc_func_name << " found!" << std::endl;
                    } else {
                        tile.does_random_ticc = true;
                        tile.random_ticc_func = ticc_funcs.at(ticc_func_name);
                    }
                }


                if (tile_entry.contains("pallet") && tile_entry["pallet"].array->size() > 0) {
                    std::vector<Color4> pallet;
                    auto pallet_meta = tile_entry["pallet"].as_array();

                    for (auto& c : pallet_meta)
                        pallet.push_back(JsonConversions::parse_color(c));

                    tile.pallet = pallet;
                }

                Tiles().Register(tile);
            }
        }
        return true;
    }

    bool LoadTileMetadata() {
        if (!LoadTileMetadata("assets/data/tiles.json")) return false;

        return true;
    }
}